# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, NamedTuple, Optional, Union

import fire
from google.protobuf import json_format

from kfp.deprecated.containers import entrypoint_utils
from kfp.deprecated.dsl import artifact
from kfp.pipeline_spec import pipeline_spec_pb2

FN_SOURCE = 'ml/main.py'
FN_NAME_ARG = 'function_name'

PARAM_METADATA_SUFFIX = '_input_param_metadata_file'
ARTIFACT_METADATA_SUFFIX = '_input_artifact_metadata_file'
FIELD_NAME_SUFFIX = '_input_field_name'
ARGO_PARAM_SUFFIX = '_input_argo_param'
INPUT_URI_SUFFIX = '_input_uri'
PRODUCER_POD_ID_SUFFIX = '_pod_id'
OUTPUT_NAME_SUFFIX = '_input_output_name'

OUTPUT_PARAM_PATH_SUFFIX = '_parameter_output_path'
OUTPUT_ARTIFACT_PATH_SUFFIX = '_artifact_output_uri'

METADATA_FILE_ARG = 'executor_metadata_json_file'


class InputParam(object):
    """POD that holds an input parameter."""

    def __init__(self,
                 value: Optional[Union[str, float, int]] = None,
                 metadata_file: Optional[str] = None,
                 field_name: Optional[str] = None):
        """Instantiates an InputParam object.

        Args:
          value: The actual value of the parameter.
          metadata_file: The location of the metadata JSON file output by the
            producer step.
          field_name: The output name of the producer.

        Raises:
          ValueError: when neither of the following is true:
            1) value is provided, and metadata_file and field_name are not; or
            2) both metadata_file and field_name are provided, and value is not.
        """
        if not (value is not None and not (metadata_file or field_name) or
                (metadata_file and field_name and value is None)):
            raise ValueError(
                'Either value or both metadata_file and field_name '
                'needs to be provided. Got value={value}, field_name='
                '{field_name}, metadata_file={metadata_file}'.format(
                    value=value,
                    field_name=field_name,
                    metadata_file=metadata_file))
        if value is not None:
            self._value = value
        else:
            # Parse the value by inspecting the producer's metadata JSON file.
            self._value = entrypoint_utils.get_parameter_from_output(
                metadata_file, field_name)

        self._metadata_file = metadata_file
        self._field_name = field_name

    # Following properties are read-only
    @property
    def value(self) -> Union[float, str, int]:
        return self._value

    @property
    def metadata_file(self) -> str:
        return self._metadata_file

    @property
    def field_name(self) -> str:
        return self._field_name


class InputArtifact(object):
    """POD that holds an input artifact."""

    def __init__(self,
                 uri: Optional[str] = None,
                 metadata_file: Optional[str] = None,
                 output_name: Optional[str] = None):
        """Instantiates an InputParam object.

        Args:
          uri: The uri holds the input artifact.
          metadata_file: The location of the metadata JSON file output by the
            producer step.
          output_name: The output name of the artifact in producer step.

        Raises:
          ValueError: when neither of the following is true:
            1) uri is provided, and metadata_file and output_name are not; or
            2) both metadata_file and output_name are provided, and uri is not.
        """
        if not ((uri and not (metadata_file or output_name) or
                 (metadata_file and output_name and not uri))):
            raise ValueError(
                'Either uri or both metadata_file and output_name '
                'needs to be provided. Got uri={uri}, output_name='
                '{output_name}, metadata_file={metadata_file}'.format(
                    uri=uri,
                    output_name=output_name,
                    metadata_file=metadata_file))

        self._metadata_file = metadata_file
        self._output_name = output_name
        if uri:
            self._uri = uri
        else:
            self._uri = self.get_artifact().uri

    # Following properties are read-only.
    @property
    def uri(self) -> str:
        return self._uri

    @property
    def metadata_file(self) -> str:
        return self._metadata_file

    @property
    def output_name(self) -> str:
        return self._output_name

    def get_artifact(self) -> artifact.Artifact:
        """Gets an artifact object by parsing metadata or creating one from
        uri."""
        if self.metadata_file and self.output_name:
            return entrypoint_utils.get_artifact_from_output(
                self.metadata_file, self.output_name)
        else:
            # Provide an empty schema when returning a raw Artifact.
            result = artifact.Artifact(
                instance_schema=artifact.DEFAULT_ARTIFACT_SCHEMA)
            result.uri = self.uri
            return result


def _write_output_metadata_file(fn_res: Union[int, str, float, NamedTuple],
                                output_artifacts: Dict[str, artifact.Artifact],
                                output_metadata_path: str):
    """Writes the output metadata file to the designated place."""
    # If output_params is a singleton value, needs to transform it to a mapping.
    output_parameters = {}
    if isinstance(fn_res, (int, str, float)):
        output_parameters['output'] = fn_res
    else:
        # When multiple outputs, we'll need to match each field to the output paths.
        for idx, output_name in enumerate(fn_res._fields):
            output_parameters[output_name] = fn_res[idx]

    executor_output = entrypoint_utils.get_executor_output(
        output_artifacts=output_artifacts, output_params=output_parameters)

    with open(output_metadata_path, 'w') as f:
        f.write(json_format.MessageToJson(executor_output))

    return executor_output


def main(executor_input_str: str,
         function_name: str,
         output_metadata_path: Optional[str] = None):
    """Container entrypoint used by KFP Python function based component.

    executor_input_str: A serialized ExecutorInput proto message.
    function_name: The name of the user-defined function.
    output_metadata_path: A local path where the output metadata JSON file should
      be written to.
    """
    executor_input = pipeline_spec_pb2.ExecutorInput()
    json_format.Parse(text=executor_input_str, message=executor_input)
    output_metadata_path = output_metadata_path or executor_input.outputs.output_file
    parameter_dict = {}  # kwargs to be passed to UDF.
    for name, input_param in executor_input.inputs.parameters.items():
        parameter_dict[name] = entrypoint_utils.get_python_value(input_param)

    for name, input_artifacts in executor_input.inputs.artifacts.items():
        parameter_dict[name] = artifact.Artifact.get_from_runtime_artifact(
            input_artifacts.artifacts[0])

    # Also, determine a way to inspect the function signature to decide the type
    # of output artifacts.
    fn = entrypoint_utils.import_func_from_source(FN_SOURCE, function_name)

    # In the ExeuctorInput message passed into the entrypoint, the output artifact
    # URIs are already specified. The output artifact is constructed according to
    # the specified URIs + type information retrieved from function signature.
    output_uris = {}
    for name, output_artifacts in executor_input.outputs.artifacts.items():
        output_uris[name] = output_artifacts.artifacts[0].uri

    output_artifacts = entrypoint_utils.get_output_artifacts(fn, output_uris)
    for name, art in output_artifacts.items():
        parameter_dict[name] = art

    # Execute the user function. fn_res is expected to contain output parameters
    # only. It's either an namedtuple or a single primitive value.
    fn_res = fn(**parameter_dict)

    _write_output_metadata_file(
        fn_res=fn_res,
        output_artifacts=output_artifacts,
        output_metadata_path=output_metadata_path)


if __name__ == '__main__':
    fire.Fire(main)



# Copyright 2022 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import unittest

from kfp import dsl
from kfp.components import importer_node
from kfp.components.types.artifact_types import Dataset


class TestImporterSupportsDynamicMetadata(unittest.TestCase):

    def test_dynamic_dict_element_from_pipeline_param(self):

        @dsl.pipeline()
        def my_pipeline(meta_inp: str):

            dataset = importer_node.importer(
                'gs://my_bucket',
                Dataset,
                reimport=True,
                metadata={
                    'string': meta_inp,
                    'string-2': meta_inp
                })

        pipeline_spec = my_pipeline.pipeline_spec
        input_keys = list(pipeline_spec.components['comp-importer']
                          .input_definitions.parameters.keys())
        self.assertIn('metadata', input_keys)
        deployment_spec = pipeline_spec.deployment_spec.fields[
            'executors'].struct_value.fields['exec-importer']
        metadata = deployment_spec.struct_value.fields[
            'importer'].struct_value.fields['metadata']
        self.assertEqual(metadata.struct_value.fields['string'].string_value,
                         "{{$.inputs.parameters['metadata']}}")
        self.assertEqual(metadata.struct_value.fields['string-2'].string_value,
                         "{{$.inputs.parameters['metadata']}}")

    def test_dynamic_list_element_from_pipeline_param(self):

        @dsl.pipeline()
        def my_pipeline(meta_inp1: str, meta_inp2: int):

            dataset = importer_node.importer(
                'gs://my_bucket',
                Dataset,
                reimport=True,
                metadata={
                    'outer_key': [meta_inp1, meta_inp2],
                    meta_inp1: meta_inp1
                })

        pipeline_spec = my_pipeline.pipeline_spec
        input_keys = list(pipeline_spec.components['comp-importer']
                          .input_definitions.parameters.keys())
        self.assertIn('metadata', input_keys)
        deployment_spec = pipeline_spec.deployment_spec.fields[
            'executors'].struct_value.fields['exec-importer']
        metadata = deployment_spec.struct_value.fields[
            'importer'].struct_value.fields['metadata']
        self.assertEqual(
            metadata.struct_value.fields['outer_key'].list_value.values[0]
            .string_value, "{{$.inputs.parameters['metadata']}}")
        self.assertEqual(
            metadata.struct_value.fields['outer_key'].list_value.values[1]
            .string_value, "{{$.inputs.parameters['metadata-2']}}")
        self.assertEqual(
            metadata.struct_value.fields["{{$.inputs.parameters['metadata']}}"]
            .string_value, "{{$.inputs.parameters['metadata']}}")

    def test_dynamic_dict_element_from_task_output(self):

        @dsl.component
        def string_task(string: str) -> str:
            return 'string'

        @dsl.pipeline()
        def my_pipeline():
            task1 = string_task(string='string1')
            task2 = string_task(string='string2')
            dataset = importer_node.importer(
                'gs://my_bucket',
                Dataset,
                reimport=True,
                metadata={
                    'string-1': task1.output,
                    'string-2': task2.output
                })

        pipeline_spec = my_pipeline.pipeline_spec
        input_keys = list(pipeline_spec.components['comp-importer']
                          .input_definitions.parameters.keys())
        self.assertIn('metadata', input_keys)
        self.assertIn('metadata-2', input_keys)

        deployment_spec = pipeline_spec.deployment_spec.fields[
            'executors'].struct_value.fields['exec-importer']
        metadata = deployment_spec.struct_value.fields[
            'importer'].struct_value.fields['metadata']
        self.assertEqual(metadata.struct_value.fields['string-1'].string_value,
                         "{{$.inputs.parameters['metadata']}}")
        self.assertEqual(metadata.struct_value.fields['string-2'].string_value,
                         "{{$.inputs.parameters['metadata-2']}}")

    def test_dynamic_list_element_from_task_output(self):

        @dsl.component
        def string_task() -> str:
            return 'string'

        @dsl.pipeline()
        def my_pipeline():
            task = string_task()
            dataset = importer_node.importer(
                'gs://my_bucket',
                Dataset,
                reimport=True,
                metadata={
                    'outer_key': [task.output, task.output],
                    task.output: task.output
                })

        pipeline_spec = my_pipeline.pipeline_spec
        input_keys = list(pipeline_spec.components['comp-importer']
                          .input_definitions.parameters.keys())
        self.assertIn('metadata', input_keys)
        self.assertNotIn('metadata-2', input_keys)
        deployment_spec = pipeline_spec.deployment_spec.fields[
            'executors'].struct_value.fields['exec-importer']
        metadata = deployment_spec.struct_value.fields[
            'importer'].struct_value.fields['metadata']
        self.assertEqual(
            metadata.struct_value.fields['outer_key'].list_value.values[0]
            .string_value, "{{$.inputs.parameters['metadata']}}")
        self.assertEqual(
            metadata.struct_value.fields['outer_key'].list_value.values[1]
            .string_value, "{{$.inputs.parameters['metadata']}}")
        self.assertEqual(
            metadata.struct_value.fields["{{$.inputs.parameters['metadata']}}"]
            .string_value, "{{$.inputs.parameters['metadata']}}")

    def test_dynamic_fstring_in_metadata(self):

        @dsl.component
        def string_task() -> str:
            return 'string'

        @dsl.pipeline()
        def my_pipeline(integer: int = 1):
            task = string_task()
            dataset = importer_node.importer(
                'gs://my_bucket',
                Dataset,
                reimport=True,
                metadata={
                    f'prefix1-{integer}': f'prefix2-{task.output}',
                    'key': 'value'
                })

        pipeline_spec = my_pipeline.pipeline_spec
        input_keys = list(pipeline_spec.components['comp-importer']
                          .input_definitions.parameters.keys())
        self.assertIn('metadata', input_keys)
        self.assertIn('metadata-2', input_keys)
        self.assertNotIn('metadata-3', input_keys)
        deployment_spec = pipeline_spec.deployment_spec.fields[
            'executors'].struct_value.fields['exec-importer']
        metadata = deployment_spec.struct_value.fields[
            'importer'].struct_value.fields['metadata']
        self.assertEqual(
            metadata.struct_value.fields[
                "prefix1-{{$.inputs.parameters[\'metadata\']}}"].string_value,
            "prefix2-{{$.inputs.parameters[\'metadata-2\']}}")
        self.assertEqual(metadata.struct_value.fields['key'].string_value,
                         'value')



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Base class for MLMD artifact in KFP SDK."""

from typing import Any, Optional

from absl import logging
import importlib
import yaml

from google.protobuf import json_format
from google.protobuf import struct_pb2
from kfp.pipeline_spec import pipeline_spec_pb2
from kfp.deprecated.dsl import serialization_utils
from kfp.deprecated.dsl import artifact_utils

KFP_ARTIFACT_ONTOLOGY_MODULE = 'kfp.dsl.ontology_artifacts'
DEFAULT_ARTIFACT_SCHEMA = 'title: kfp.Artifact\ntype: object\nproperties:\n'


class Artifact(object):
    """KFP Artifact Python class.

    Artifact Python class/object mainly serves following purposes in different
    period of its lifecycle.

    1. During compile time, users can use Artifact class to annotate I/O types of
       their components.
    2. At runtime, Artifact objects provide helper function/utilities to access
       the underlying RuntimeArtifact pb message, and provide additional layers
       of validation to ensure type compatibility for fields specified in the
       instance schema.
    """

    TYPE_NAME = "kfp.Artifact"

    # Initialization flag to support setattr / getattr behavior.
    _initialized = False

    def __init__(self, instance_schema: Optional[str] = None):
        """Constructs an instance of Artifact.

        Setups up self._metadata_fields to perform type checking and
        initialize RuntimeArtifact.
        """
        if self.__class__ == Artifact:
            if not instance_schema:
                raise ValueError(
                    'The "instance_schema" argument must be set for Artifact.')
            self._instance_schema = instance_schema
        else:
            if instance_schema:
                raise ValueError(
                    'The "instance_schema" argument must not be passed for Artifact \
               subclass: {}'.format(self.__class__))

        # setup self._metadata_fields
        self.TYPE_NAME, self._metadata_fields = artifact_utils.parse_schema(
            self._instance_schema)

        # Instantiate a RuntimeArtifact pb message as the POD data structure.
        self._artifact = pipeline_spec_pb2.RuntimeArtifact()

        # Stores the metadata for the Artifact.
        self.metadata = {}

        self._artifact.type.CopyFrom(
            pipeline_spec_pb2.ArtifactTypeSchema(
                instance_schema=self._instance_schema))

        self._initialized = True

    @property
    def type_schema(self) -> str:
        """Gets the instance_schema for this Artifact object."""
        return self._instance_schema

    def __getattr__(self, name: str) -> Any:
        """Custom __getattr__ to allow access to artifact metadata."""

        if name not in self._metadata_fields:
            raise AttributeError(
                'No metadata field: {} in artifact.'.format(name))

        return self.metadata[name]

    def __setattr__(self, name: str, value: Any):
        """Custom __setattr__ to allow access to artifact metadata."""

        if not self._initialized:
            object.__setattr__(self, name, value)
            return

        metadata_fields = {}
        if self._metadata_fields:
            metadata_fields = self._metadata_fields

        if name not in self._metadata_fields:
            if (name in self.__dict__ or
                    any(name in c.__dict__ for c in self.__class__.mro())):
                # Use any provided getter / setter if available.
                object.__setattr__(self, name, value)
                return
            # In the case where we do not handle this via an explicit getter /
            # setter, we assume that the user implied an artifact attribute store,
            # and we raise an exception since such an attribute was not explicitly
            # defined in the Artifact PROPERTIES dictionary.
            raise AttributeError('Cannot set an unspecified metadata field:{} \
        on artifact. Only fields specified in instance schema can be \
        set.'.format(name))

        # Type checking to be performed during serialization.
        self.metadata[name] = value

    def _update_runtime_artifact(self):
        """Verifies metadata is well-formed and updates artifact instance."""

        artifact_utils.verify_schema_instance(self._instance_schema,
                                              self.metadata)

        if len(self.metadata) != 0:
            metadata_protobuf_struct = struct_pb2.Struct()
            metadata_protobuf_struct.update(self.metadata)
            self._artifact.metadata.CopyFrom(metadata_protobuf_struct)

    @property
    def type(self):
        return self.__class__

    @property
    def type_name(self):
        return self.TYPE_NAME

    @property
    def uri(self) -> str:
        return self._artifact.uri

    @uri.setter
    def uri(self, uri: str) -> None:
        self._artifact.uri = uri

    @property
    def name(self) -> str:
        return self._artifact.name

    @name.setter
    def name(self, name: str) -> None:
        self._artifact.name = name

    @property
    def runtime_artifact(self) -> pipeline_spec_pb2.RuntimeArtifact:
        self._update_runtime_artifact()
        return self._artifact

    @runtime_artifact.setter
    def runtime_artifact(self, artifact: pipeline_spec_pb2.RuntimeArtifact):
        self._artifact = artifact

    def serialize(self) -> str:
        """Serializes an Artifact to JSON dict format."""
        self._update_runtime_artifact()
        return json_format.MessageToJson(self._artifact, sort_keys=True)

    @classmethod
    def get_artifact_type(cls) -> str:
        """Gets the instance_schema according to the Python schema spec."""
        result_map = {'title': cls.TYPE_NAME, 'type': 'object'}

        return serialization_utils.yaml_dump(result_map)

    @classmethod
    def get_ir_type(cls) -> pipeline_spec_pb2.ArtifactTypeSchema:
        return pipeline_spec_pb2.ArtifactTypeSchema(
            instance_schema=cls.get_artifact_type())

    @classmethod
    def get_from_runtime_artifact(
            cls, artifact: pipeline_spec_pb2.RuntimeArtifact) -> Any:
        """Deserializes an Artifact object from RuntimeArtifact message."""
        instance_schema = yaml.safe_load(artifact.type.instance_schema)
        type_name = instance_schema['title'][len('kfp.'):]
        result = None
        try:
            artifact_cls = getattr(
                importlib.import_module(KFP_ARTIFACT_ONTOLOGY_MODULE),
                type_name)
            result = artifact_cls()
        except (AttributeError, ImportError, ValueError) as err:
            logging.warning('Failed to instantiate Ontology Artifact:{} \
        instance'.format(type_name))

        if not result:
            # Otherwise generate a generic Artifact object.
            result = Artifact(instance_schema=artifact.type.instance_schema)
        result.runtime_artifact = artifact
        result.metadata = json_format.MessageToDict(artifact.metadata)
        return result

    @classmethod
    def deserialize(cls, data: str) -> Any:
        """Deserializes an Artifact object from JSON dict."""
        artifact = pipeline_spec_pb2.RuntimeArtifact()
        json_format.Parse(data, artifact, ignore_unknown_fields=True)
        return cls.get_from_runtime_artifact(artifact)



# Copyright 2019 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import warnings
from typing import (Any, Callable, Dict, List, Optional, Sequence, Tuple,
                    TypeVar, Union)

from kfp.deprecated._config import COMPILING_FOR_V2
from kfp.deprecated.components import _components, _structures
from kfp.deprecated.dsl import _pipeline_param
from kfp.pipeline_spec import pipeline_spec_pb2
from kubernetes.client import V1Affinity, V1Toleration
from kubernetes.client.models import (V1Container, V1ContainerPort,
                                      V1EnvFromSource, V1EnvVar, V1Lifecycle,
                                      V1Probe, V1ResourceRequirements,
                                      V1SecurityContext, V1Volume,
                                      V1VolumeDevice, V1VolumeMount)

# generics
T = TypeVar('T')
# type alias: either a string or a list of string
StringOrStringList = Union[str, List[str]]
ContainerOpArgument = Union[str, int, float, bool,
                            _pipeline_param.PipelineParam]
ArgumentOrArguments = Union[ContainerOpArgument, List]

ALLOWED_RETRY_POLICIES = (
    'Always',
    'OnError',
    'OnFailure',
    'OnTransientError',
)

# Shorthand for PipelineContainerSpec
_PipelineContainerSpec = pipeline_spec_pb2.PipelineDeploymentConfig.PipelineContainerSpec

# Unit constants for k8s size string.
_E = 10**18  # Exa
_EI = 1 << 60  # Exa: power-of-two approximate
_P = 10**15  # Peta
_PI = 1 << 50  # Peta: power-of-two approximate
# noinspection PyShadowingBuiltins
_T = 10**12  # Tera
_TI = 1 << 40  # Tera: power-of-two approximate
_G = 10**9  # Giga
_GI = 1 << 30  # Giga: power-of-two approximate
_M = 10**6  # Mega
_MI = 1 << 20  # Mega: power-of-two approximate
_K = 10**3  # Kilo
_KI = 1 << 10  # Kilo: power-of-two approximate

_GKE_ACCELERATOR_LABEL = 'cloud.google.com/gke-accelerator'

_DEFAULT_CUSTOM_JOB_MACHINE_TYPE = 'n1-standard-4'


# util functions
def deprecation_warning(func: Callable, op_name: str,
                        container_name: str) -> Callable:
    """Decorator function to give a pending deprecation warning."""

    def _wrapped(*args, **kwargs):
        warnings.warn(
            '`dsl.ContainerOp.%s` will be removed in future releases. '
            'Use `dsl.ContainerOp.container.%s` instead.' %
            (op_name, container_name), PendingDeprecationWarning)
        return func(*args, **kwargs)

    return _wrapped


def _create_getter_setter(prop):
    """Create a tuple of getter and setter methods for a property in
    `Container`."""

    def _getter(self):
        return getattr(self._container, prop)

    def _setter(self, value):
        return setattr(self._container, prop, value)

    return _getter, _setter


def _proxy_container_op_props(cls: 'ContainerOp'):
    """Takes the `ContainerOp` class and proxy the PendingDeprecation
    properties in `ContainerOp` to the `Container` instance."""
    # properties mapping to proxy: ContainerOps.<prop> => Container.<prop>
    prop_map = dict(image='image', env_variables='env')
    # itera and create class props
    for op_prop, container_prop in prop_map.items():
        # create getter and setter
        _getter, _setter = _create_getter_setter(container_prop)
        # decorate with deprecation warning
        getter = deprecation_warning(_getter, op_prop, container_prop)
        setter = deprecation_warning(_setter, op_prop, container_prop)
        # update attribites with properties
        setattr(cls, op_prop, property(getter, setter))
    return cls


def as_string_list(
        list_or_str: Optional[Union[Any, Sequence[Any]]]) -> List[str]:
    """Convert any value except None to a list if not already a list."""
    if list_or_str is None:
        return None
    if isinstance(list_or_str, Sequence) and not isinstance(list_or_str, str):
        list_value = list_or_str
    else:
        list_value = [list_or_str]
    return [str(item) for item in list_value]


def create_and_append(current_list: Union[List[T], None], item: T) -> List[T]:
    """Create a list (if needed) and appends an item to it."""
    current_list = current_list or []
    current_list.append(item)
    return current_list


class Container(V1Container):
    """A wrapper over k8s container definition object
    (io.k8s.api.core.v1.Container), which is used to represent the `container`
    property in argo's workflow template
    (io.argoproj.workflow.v1alpha1.Template).

    `Container` class also comes with utility functions to set and update the
    the various properties for a k8s container definition.

    NOTE: A notable difference is that `name` is not required and will not be
    processed for `Container` (in contrast to `V1Container` where `name` is a
    required property).

    See:
    *
    https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_container.py
    * https://github.com/argoproj/argo-workflows/blob/master/api/openapi-spec/swagger.json

    Example::

      from kfp.dsl import ContainerOp
      from kubernetes.client.models import V1EnvVar


      # creates a operation
      op = ContainerOp(name='bash-ops',
                      image='busybox:latest',
                      command=['echo'],
                      arguments=['$MSG'])

      # returns a `Container` object from `ContainerOp`
      # and add an environment variable to `Container`
      op.container.add_env_variable(V1EnvVar(name='MSG', value='hello world'))

    Attributes:
      attribute_map (dict): The key is attribute name and the value is json key
        in definition.
    """
    # remove `name` from attribute_map, swagger_types and openapi_types so `name` is not generated in the JSON

    if hasattr(V1Container, 'swagger_types'):
        swagger_types = {
            key: value
            for key, value in V1Container.swagger_types.items()
            if key != 'name'
        }
    if hasattr(V1Container, 'openapi_types'):
        openapi_types = {
            key: value
            for key, value in V1Container.openapi_types.items()
            if key != 'name'
        }
    attribute_map = {
        key: value
        for key, value in V1Container.attribute_map.items()
        if key != 'name'
    }

    def __init__(self, image: str, command: List[str], args: List[str],
                 **kwargs):
        """Creates a new instance of `Container`.

        Args:
          image {str}: image to use, e.g. busybox:latest
          command {List[str]}: entrypoint array.  Not executed within a shell.
          args {List[str]}: arguments to entrypoint.
          **kwargs: keyword arguments for `V1Container`
        """
        # set name to '' if name is not provided
        # k8s container MUST have a name
        # argo workflow template does not need a name for container def
        if not kwargs.get('name'):
            kwargs['name'] = ''

        # v2 container_spec
        self._container_spec = None

        self.env_dict = {}

        super(Container, self).__init__(
            image=image, command=command, args=args, **kwargs)

    def _validate_size_string(self, size_string):
        """Validate a given string is valid for memory/ephemeral-storage
        request or limit."""

        if isinstance(size_string, _pipeline_param.PipelineParam):
            if size_string.value:
                size_string = size_string.value
            else:
                return

        if re.match(r'^[0-9]+(E|Ei|P|Pi|T|Ti|G|Gi|M|Mi|K|Ki){0,1}$',
                    size_string) is None:
            raise ValueError(
                'Invalid memory string. Should be an integer, or integer followed '
                'by one of "E|Ei|P|Pi|T|Ti|G|Gi|M|Mi|K|Ki"')

    def _validate_cpu_string(self, cpu_string):
        """Validate a given string is valid for cpu request or limit."""

        if isinstance(cpu_string, _pipeline_param.PipelineParam):
            if cpu_string.value:
                cpu_string = cpu_string.value
            else:
                return

        if re.match(r'^[0-9]+m$', cpu_string) is not None:
            return

        try:
            float(cpu_string)
        except ValueError:
            raise ValueError(
                'Invalid cpu string. Should be float or integer, or integer followed '
                'by "m".')

    def _validate_positive_number(self, str_value, param_name):
        """Validate a given string is in positive integer format."""

        if isinstance(str_value, _pipeline_param.PipelineParam):
            if str_value.value:
                str_value = str_value.value
            else:
                return

        try:
            int_value = int(str_value)
        except ValueError:
            raise ValueError(
                'Invalid {}. Should be integer.'.format(param_name))

        if int_value <= 0:
            raise ValueError('{} must be positive integer.'.format(param_name))

    def add_resource_limit(self, resource_name, value) -> 'Container':
        """Add the resource limit of the container.

        Args:
          resource_name: The name of the resource. It can be cpu, memory, etc.
          value: The string value of the limit.
        """

        self.resources = self.resources or V1ResourceRequirements()
        self.resources.limits = self.resources.limits or {}
        self.resources.limits.update({resource_name: value})
        return self

    def add_resource_request(self, resource_name, value) -> 'Container':
        """Add the resource request of the container.

        Args:
          resource_name: The name of the resource. It can be cpu, memory, etc.
          value: The string value of the request.
        """

        self.resources = self.resources or V1ResourceRequirements()
        self.resources.requests = self.resources.requests or {}
        self.resources.requests.update({resource_name: value})
        return self

    def get_resource_limit(self, resource_name: str) -> Optional[str]:
        """Get the resource limit of the container.

        Args:
          resource_name: The name of the resource. It can be cpu, memory, etc.
        """

        if not self.resources or not self.resources.limits:
            return None
        return self.resources.limits.get(resource_name)

    def get_resource_request(self, resource_name: str) -> Optional[str]:
        """Get the resource request of the container.

        Args:
          resource_name: The name of the resource. It can be cpu, memory, etc.
        """

        if not self.resources or not self.resources.requests:
            return None
        return self.resources.requests.get(resource_name)

    def set_memory_request(
            self, memory: Union[str,
                                _pipeline_param.PipelineParam]) -> 'Container':
        """Set memory request (minimum) for this operator.

        Args:
          memory(Union[str, PipelineParam]): a string which can be a number or a number followed by one of
            "E", "P", "T", "G", "M", "K".
        """

        if not isinstance(memory, _pipeline_param.PipelineParam):
            self._validate_size_string(memory)
        return self.add_resource_request('memory', memory)

    def set_memory_limit(
            self, memory: Union[str,
                                _pipeline_param.PipelineParam]) -> 'Container':
        """Set memory limit (maximum) for this operator.

        Args:
          memory(Union[str, PipelineParam]): a string which can be a number or a number followed by one of
            "E", "P", "T", "G", "M", "K".
        """
        if not isinstance(memory, _pipeline_param.PipelineParam):
            self._validate_size_string(memory)
            if self._container_spec:
                self._container_spec.resources.memory_limit = _get_resource_number(
                    memory)
        return self.add_resource_limit('memory', memory)

    def set_ephemeral_storage_request(self, size) -> 'Container':
        """Set ephemeral-storage request (minimum) for this operator.

        Args:
          size: a string which can be a number or a number followed by one of
            "E", "P", "T", "G", "M", "K".
        """
        self._validate_size_string(size)
        return self.add_resource_request('ephemeral-storage', size)

    def set_ephemeral_storage_limit(self, size) -> 'Container':
        """Set ephemeral-storage request (maximum) for this operator.

        Args:
          size: a string which can be a number or a number followed by one of
            "E", "P", "T", "G", "M", "K".
        """
        self._validate_size_string(size)
        return self.add_resource_limit('ephemeral-storage', size)

    def set_cpu_request(
            self, cpu: Union[str,
                             _pipeline_param.PipelineParam]) -> 'Container':
        """Set cpu request (minimum) for this operator.

        Args:
          cpu(Union[str, PipelineParam]): A string which can be a number or a number followed by "m", which
            means 1/1000.
        """
        if not isinstance(cpu, _pipeline_param.PipelineParam):
            self._validate_cpu_string(cpu)
        return self.add_resource_request('cpu', cpu)

    def set_cpu_limit(
            self, cpu: Union[str,
                             _pipeline_param.PipelineParam]) -> 'Container':
        """Set cpu limit (maximum) for this operator.

        Args:
          cpu(Union[str, PipelineParam]): A string which can be a number or a number followed by "m", which
            means 1/1000.
        """

        if not isinstance(cpu, _pipeline_param.PipelineParam):
            self._validate_cpu_string(cpu)
            if self._container_spec:
                self._container_spec.resources.cpu_limit = _get_cpu_number(cpu)
        return self.add_resource_limit('cpu', cpu)

    def set_gpu_limit(
        self,
        gpu: Union[str, _pipeline_param.PipelineParam],
        vendor: Union[str, _pipeline_param.PipelineParam] = 'nvidia'
    ) -> 'Container':
        """Set gpu limit for the operator.

        This function add '<vendor>.com/gpu' into resource limit.
        Note that there is no need to add GPU request. GPUs are only supposed to
        be specified in the limits section. See
        https://kubernetes.io/docs/tasks/manage-gpus/scheduling-gpus/.

        Args:
          gpu(Union[str, PipelineParam]): A string which must be a positive number.
          vendor(Union[str, PipelineParam]): Optional. A string which is the vendor of the requested gpu.
            The supported values are: 'nvidia' (default), and 'amd'. The value is
            ignored in v2.
        """

        if not isinstance(gpu, _pipeline_param.PipelineParam):
            self._validate_positive_number(gpu, 'gpu')

            if self._container_spec:
                # For backforward compatibiliy, allow `gpu` to be a string.
                self._container_spec.resources.accelerator.count = int(gpu)

            if vendor != 'nvidia' and vendor != 'amd':
                raise ValueError('vendor can only be nvidia or amd.')

            return self.add_resource_limit('%s.com/gpu' % vendor, gpu)

        return self.add_resource_limit(vendor, gpu)

    def add_volume_mount(self, volume_mount) -> 'Container':
        """Add volume to the container.

        Args:
          volume_mount: Kubernetes volume mount For detailed spec, check volume
            mount definition
          https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_volume_mount.py
        """

        if not isinstance(volume_mount, V1VolumeMount):
            raise ValueError(
                'invalid argument. Must be of instance `V1VolumeMount`.')

        self.volume_mounts = create_and_append(self.volume_mounts, volume_mount)
        return self

    def add_volume_devices(self, volume_device) -> 'Container':
        """Add a block device to be used by the container.

        Args:
          volume_device: Kubernetes volume device For detailed spec, volume
            device definition
          https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_volume_device.py
        """

        if not isinstance(volume_device, V1VolumeDevice):
            raise ValueError(
                'invalid argument. Must be of instance `V1VolumeDevice`.')

        self.volume_devices = create_and_append(self.volume_devices,
                                                volume_device)
        return self

    def set_env_variable(self, name: str, value: str) -> 'Container':
        """Sets environment variable to the container (v2 only).

        Args:
          name: The name of the environment variable.
          value: The value of the environment variable.
        """

        if not COMPILING_FOR_V2:
            raise ValueError(
                'set_env_variable is v2 only. Use add_env_variable for v1.')

        # Merge with any existing environment varaibles
        self.env_dict = {
            env.name: env.value for env in self._container_spec.env or []
        }
        self.env_dict[name] = value

        del self._container_spec.env[:]
        self._container_spec.env.extend([
            _PipelineContainerSpec.EnvVar(name=name, value=value)
            for name, value in self.env_dict.items()
        ])
        return self

    def add_env_variable(self, env_variable) -> 'Container':
        """Add environment variable to the container.

        Args:
          env_variable: Kubernetes environment variable For detailed spec, check
            environment variable definition
          https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_env_var.py
        """

        if not isinstance(env_variable, V1EnvVar):
            raise ValueError(
                'invalid argument. Must be of instance `V1EnvVar`.')

        self.env = create_and_append(self.env, env_variable)
        return self

    def add_env_from(self, env_from) -> 'Container':
        """Add a source to populate environment variables int the container.

        Args:
          env_from: Kubernetes environment from source For detailed spec, check
            environment from source definition
          https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_env_var_source.py
        """

        if not isinstance(env_from, V1EnvFromSource):
            raise ValueError(
                'invalid argument. Must be of instance `V1EnvFromSource`.')

        self.env_from = create_and_append(self.env_from, env_from)
        return self

    def set_image_pull_policy(self, image_pull_policy) -> 'Container':
        """Set image pull policy for the container.

        Args:
          image_pull_policy: One of `Always`, `Never`, `IfNotPresent`.
        """
        if image_pull_policy not in ['Always', 'Never', 'IfNotPresent']:
            raise ValueError(
                'Invalid imagePullPolicy. Must be one of `Always`, `Never`, `IfNotPresent`.'
            )

        self.image_pull_policy = image_pull_policy
        return self

    def add_port(self, container_port) -> 'Container':
        """Add a container port to the container.

        Args:
          container_port: Kubernetes container port For detailed spec, check
            container port definition
          https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_container_port.py
        """

        if not isinstance(container_port, V1ContainerPort):
            raise ValueError(
                'invalid argument. Must be of instance `V1ContainerPort`.')

        self.ports = create_and_append(self.ports, container_port)
        return self

    def set_security_context(self, security_context) -> 'Container':
        """Set security configuration to be applied on the container.

        Args:
          security_context: Kubernetes security context For detailed spec, check
            security context definition
          https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_security_context.py
        """

        if not isinstance(security_context, V1SecurityContext):
            raise ValueError(
                'invalid argument. Must be of instance `V1SecurityContext`.')

        self.security_context = security_context
        return self

    def set_stdin(self, stdin=True) -> 'Container':
        """Whether this container should allocate a buffer for stdin in the
        container runtime. If this is not set, reads from stdin in the
        container will always result in EOF.

        Args:
          stdin: boolean flag
        """

        self.stdin = stdin
        return self

    def set_stdin_once(self, stdin_once=True) -> 'Container':
        """Whether the container runtime should close the stdin channel after
        it has been opened by a single attach. When stdin is true the stdin
        stream will remain open across multiple attach sessions. If stdinOnce
        is set to true, stdin is opened on container start, is empty until the
        first client attaches to stdin, and then remains open and accepts data
        until the client disconnects, at which time stdin is closed and remains
        closed until the container is restarted. If this flag is false, a
        container processes that reads from stdin will never receive an EOF.

        Args:
          stdin_once: boolean flag
        """

        self.stdin_once = stdin_once
        return self

    def set_termination_message_path(self,
                                     termination_message_path) -> 'Container':
        """Path at which the file to which the container's termination message
        will be written is mounted into the container's filesystem. Message
        written is intended to be brief final status, such as an assertion
        failure message. Will be truncated by the node if greater than 4096
        bytes. The total message length across all containers will be limited
        to 12kb.

        Args:
          termination_message_path: path for the termination message
        """
        self.termination_message_path = termination_message_path
        return self

    def set_termination_message_policy(
            self, termination_message_policy) -> 'Container':
        """Indicate how the termination message should be populated. File will
        use the contents of terminationMessagePath to populate the container
        status message on both success and failure. FallbackToLogsOnError will
        use the last chunk of container log output if the termination message
        file is empty and the container exited with an error. The log output is
        limited to 2048 bytes or 80 lines, whichever is smaller.

        Args:
          termination_message_policy: `File` or `FallbackToLogsOnError`
        """
        if termination_message_policy not in ['File', 'FallbackToLogsOnError']:
            raise ValueError(
                'terminationMessagePolicy must be `File` or `FallbackToLogsOnError`'
            )
        self.termination_message_policy = termination_message_policy
        return self

    def set_tty(self, tty: bool = True) -> 'Container':
        """Whether this container should allocate a TTY for itself, also
        requires 'stdin' to be true.

        Args:
          tty: boolean flag
        """

        self.tty = tty
        return self

    def set_readiness_probe(self, readiness_probe) -> 'Container':
        """Set a readiness probe for the container.

        Args:
          readiness_probe: Kubernetes readiness probe For detailed spec, check
            probe definition
          https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_probe.py
        """

        if not isinstance(readiness_probe, V1Probe):
            raise ValueError('invalid argument. Must be of instance `V1Probe`.')

        self.readiness_probe = readiness_probe
        return self

    def set_liveness_probe(self, liveness_probe) -> 'Container':
        """Set a liveness probe for the container.

        Args:
          liveness_probe: Kubernetes liveness probe For detailed spec, check
            probe definition
          https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_probe.py
        """

        if not isinstance(liveness_probe, V1Probe):
            raise ValueError('invalid argument. Must be of instance `V1Probe`.')

        self.liveness_probe = liveness_probe
        return self

    def set_lifecycle(self, lifecycle) -> 'Container':
        """Setup a lifecycle config for the container.

        Args:
          lifecycle: Kubernetes lifecycle For detailed spec, lifecycle
            definition
          https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_lifecycle.py
        """

        if not isinstance(lifecycle, V1Lifecycle):
            raise ValueError(
                'invalid argument. Must be of instance `V1Lifecycle`.')

        self.lifecycle = lifecycle
        return self


class UserContainer(Container):
    """Represents an argo workflow UserContainer
    (io.argoproj.workflow.v1alpha1.UserContainer) to be used in `UserContainer`
    property in argo's workflow template
    (io.argoproj.workflow.v1alpha1.Template).

    `UserContainer` inherits from `Container` class with an addition of
    `mirror_volume_mounts`
    attribute (`mirrorVolumeMounts` property).

    See
    https://github.com/argoproj/argo-workflows/blob/master/api/openapi-spec/swagger.json

    Args:
      name: unique name for the user container
      image: image to use for the user container, e.g. redis:alpine
      command: entrypoint array.  Not executed within a shell.
      args: arguments to the entrypoint.
      mirror_volume_mounts: MirrorVolumeMounts will mount the same volumes
        specified in the main container to the container (including
        artifacts), at the same mountPaths. This enables dind daemon to
        partially see the same filesystem as the main container in order to
        use features such as docker volume binding
      **kwargs: keyword arguments available for `Container`

    Attributes:
      swagger_types (dict): The key is attribute name and the value is attribute
        type.

    Example ::

      from kfp.dsl import ContainerOp, UserContainer
      # creates a `ContainerOp` and adds a redis init container
      op = (ContainerOp(name='foo-op', image='busybox:latest')
         .add_initContainer(UserContainer(name='redis', image='redis:alpine')))
    """
    # adds `mirror_volume_mounts` to `UserContainer` swagger definition
    # NOTE inherits definition from `V1Container` rather than `Container`
    #      because `Container` has no `name` property.
    if hasattr(V1Container, 'swagger_types'):
        swagger_types = dict(
            **V1Container.swagger_types, mirror_volume_mounts='bool')
    if hasattr(V1Container, 'openapi_types'):
        openapi_types = dict(
            **V1Container.openapi_types, mirror_volume_mounts='bool')
    attribute_map = dict(
        **V1Container.attribute_map, mirror_volume_mounts='mirrorVolumeMounts')

    def __init__(self,
                 name: str,
                 image: str,
                 command: StringOrStringList = None,
                 args: StringOrStringList = None,
                 mirror_volume_mounts: bool = None,
                 **kwargs):
        super().__init__(
            name=name,
            image=image,
            command=as_string_list(command),
            args=as_string_list(args),
            **kwargs)

        self.mirror_volume_mounts = mirror_volume_mounts

    def set_mirror_volume_mounts(self, mirror_volume_mounts=True):
        """Setting mirrorVolumeMounts to true will mount the same volumes
        specified in the main container to the container (including artifacts),
        at the same mountPaths. This enables dind daemon to partially see the
        same filesystem as the main container in order to use features such as
        docker volume binding.

        Args:
            mirror_volume_mounts: boolean flag
        """

        self.mirror_volume_mounts = mirror_volume_mounts
        return self

    @property
    def inputs(self):
        """A list of PipelineParam found in the UserContainer object."""
        return _pipeline_param.extract_pipelineparams_from_any(self)


class Sidecar(UserContainer):
    """Creates a new instance of `Sidecar`.

    Args:
      name: unique name for the sidecar container
      image: image to use for the sidecar container, e.g. redis:alpine
      command: entrypoint array.  Not executed within a shell.
      args: arguments to the entrypoint.
      mirror_volume_mounts: MirrorVolumeMounts will mount the same volumes
        specified in the main container to the sidecar (including artifacts),
        at the same mountPaths. This enables dind daemon to partially see the
        same filesystem as the main container in order to use features such as
        docker volume binding
      **kwargs: keyword arguments available for `Container`
    """

    def __init__(self,
                 name: str,
                 image: str,
                 command: StringOrStringList = None,
                 args: StringOrStringList = None,
                 mirror_volume_mounts: bool = None,
                 **kwargs):
        super().__init__(
            name=name,
            image=image,
            command=command,
            args=args,
            mirror_volume_mounts=mirror_volume_mounts,
            **kwargs)


def _make_hash_based_id_for_op(op):
    # Generating a unique ID for Op. For class instances, the hash is the object's memory address which is unique.
    return op.human_name + ' ' + hex(2**63 + hash(op))[2:]


# Pointer to a function that generates a unique ID for the Op instance (Possibly by registering the Op instance in some system).
_register_op_handler = _make_hash_based_id_for_op


class BaseOp(object):
    """Base operator.

    Args:
      name: the name of the op. It does not have to be unique within a
        pipeline because the pipeline will generates a unique new name in case
        of conflicts.
      init_containers: the list of `UserContainer` objects describing the
        InitContainer to deploy before the `main` container.
      sidecars: the list of `Sidecar` objects describing the sidecar
        containers to deploy together with the `main` container.
      is_exit_handler: Deprecated.
    """

    # list of attributes that might have pipeline params - used to generate
    # the input parameters during compilation.
    # Excludes `file_outputs` and `outputs` as they are handled separately
    # in the compilation process to generate the DAGs and task io parameters.
    attrs_with_pipelineparams = [
        'node_selector', 'volumes', 'pod_annotations', 'pod_labels',
        'num_retries', 'init_containers', 'sidecars', 'tolerations'
    ]

    def __init__(self,
                 name: str,
                 init_containers: List[UserContainer] = None,
                 sidecars: List[Sidecar] = None,
                 is_exit_handler: bool = False):

        if is_exit_handler:
            warnings.warn('is_exit_handler=True is no longer needed.',
                          DeprecationWarning)

        self.is_exit_handler = is_exit_handler

        # human_name must exist to construct operator's name
        self.human_name = name
        self.display_name = None  #TODO Set display_name to human_name
        # ID of the current Op. Ideally, it should be generated by the compiler that sees the bigger context.
        # However, the ID is used in the task output references (PipelineParams) which can be serialized to strings.
        # Because of this we must obtain a unique ID right now.
        self.name = _register_op_handler(self)

        # TODO: proper k8s definitions so that `convert_k8s_obj_to_json` can be used?
        # `io.argoproj.workflow.v1alpha1.Template` properties
        self.node_selector = {}
        self.volumes = []
        self.tolerations = []
        self.affinity = {}
        self.pod_annotations = {}
        self.pod_labels = {}

        # Retry strategy
        self.num_retries = 0
        self.retry_policy = None
        self.backoff_factor = None
        self.backoff_duration = None
        self.backoff_max_duration = None

        self.timeout = 0
        self.init_containers = init_containers or []
        self.sidecars = sidecars or []

        # used to mark this op with loop arguments
        self.loop_args = None

        # Placeholder for inputs when adding ComponentSpec metadata to this
        # ContainerOp. This holds inputs defined in ComponentSpec that have
        # a corresponding PipelineParam.
        self._component_spec_inputs_with_pipeline_params = []

        # attributes specific to `BaseOp`
        self._inputs = []
        self.dependent_names = []

        # Caching option, default to True
        self.enable_caching = True

    @property
    def inputs(self):
        """List of PipelineParams that will be converted into input parameters
        (io.argoproj.workflow.v1alpha1.Inputs) for the argo workflow."""
        # Iterate through and extract all the `PipelineParam` in Op when
        # called the 1st time (because there are in-place updates to `PipelineParam`
        # during compilation - remove in-place updates for easier debugging?)
        if not self._inputs:
            self._inputs = self._component_spec_inputs_with_pipeline_params or []
            # TODO replace with proper k8s obj?
            for key in self.attrs_with_pipelineparams:
                self._inputs += _pipeline_param.extract_pipelineparams_from_any(
                    getattr(self, key))
            # keep only unique
            self._inputs = list(set(self._inputs))
        return self._inputs

    @inputs.setter
    def inputs(self, value):
        # to support in-place updates
        self._inputs = value

    def apply(self, mod_func):
        """Applies a modifier function to self.

        The function should return the passed object.
        This is needed to chain "extention methods" to this class.

        Example::

          from kfp.gcp import use_gcp_secret
          task = (
              train_op(...)
                  .set_memory_request('1G')
                  .apply(use_gcp_secret('user-gcp-sa'))
                  .set_memory_limit('2G')
          )
        """
        return mod_func(self) or self

    def after(self, *ops):
        """Specify explicit dependency on other ops."""
        for op in ops:
            self.dependent_names.append(op.name)
        return self

    def add_volume(self, volume):
        """Add K8s volume to the container.

        Args:
          volume: Kubernetes volumes For detailed spec, check volume definition
          https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_volume.py
        """
        self.volumes.append(volume)
        return self

    def add_toleration(self, tolerations: V1Toleration):
        """Add K8s tolerations.

        Args:
          tolerations: Kubernetes toleration For detailed spec, check toleration
            definition
            https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_toleration.py
        """
        self.tolerations.append(tolerations)
        return self

    def add_affinity(self, affinity: V1Affinity):
        """Add K8s Affinity.

        Args:
          affinity: Kubernetes affinity For detailed spec, check affinity
            definition
          https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_affinity.py

        Example::

          V1Affinity(
              node_affinity=V1NodeAffinity(
                  required_during_scheduling_ignored_during_execution=V1NodeSelector(
                      node_selector_terms=[V1NodeSelectorTerm(
                          match_expressions=[V1NodeSelectorRequirement(
                              key='beta.kubernetes.io/instance-type',
                              operator='In',
                              values=['p2.xlarge'])])])))
        """
        self.affinity = affinity
        return self

    def add_node_selector_constraint(
            self, label_name: Union[str, _pipeline_param.PipelineParam],
            value: Union[str, _pipeline_param.PipelineParam]):
        """Add a constraint for nodeSelector.

        Each constraint is a key-value pair label.
        For the container to be eligible to run on a node, the node must have each
        of the constraints appeared as labels.

        Args:
          label_name(Union[str, PipelineParam]): The name of the constraint label.
          value(Union[str, PipelineParam]): The value of the constraint label.
        """

        self.node_selector[label_name] = value
        return self

    def add_pod_annotation(self, name: str, value: str):
        """Adds a pod's metadata annotation.

        Args:
          name: The name of the annotation.
          value: The value of the annotation.
        """

        self.pod_annotations[name] = value
        return self

    def add_pod_label(self, name: str, value: str):
        """Adds a pod's metadata label.

        Args:
          name: The name of the label.
          value: The value of the label.
        """

        self.pod_labels[name] = value
        return self

    def set_retry(self,
                  num_retries: int,
                  policy: Optional[str] = None,
                  backoff_duration: Optional[str] = None,
                  backoff_factor: Optional[float] = None,
                  backoff_max_duration: Optional[str] = None):
        """Sets the number of times the task is retried until it's declared
        failed.

        Args:
          num_retries: Number of times to retry on failures.
          policy: Retry policy name.
          backoff_duration: The time interval between retries. Defaults to an
            immediate retry. In case you specify a simple number, the unit
            defaults to seconds. You can also specify a different unit, for
            instance, 2m (2 minutes), 1h (1 hour).
          backoff_factor: The exponential backoff factor applied to
            backoff_duration. For example, if backoff_duration="60"
            (60 seconds) and backoff_factor=2, the first retry will happen
            after 60 seconds, then after 120, 240, and so on.
          backoff_max_duration: The maximum interval that can be reached with
            the backoff strategy.
        """
        if policy is not None and policy not in ALLOWED_RETRY_POLICIES:
            raise ValueError('policy must be one of: %r' %
                             (ALLOWED_RETRY_POLICIES,))

        self.num_retries = num_retries
        self.retry_policy = policy
        self.backoff_factor = backoff_factor
        self.backoff_duration = backoff_duration
        self.backoff_max_duration = backoff_max_duration
        return self

    def set_timeout(self, seconds: int):
        """Sets the timeout for the task in seconds.

        Args:
          seconds: Number of seconds.
        """

        self.timeout = seconds
        return self

    def add_init_container(self, init_container: UserContainer):
        """Add a init container to the Op.

        Args:
          init_container: UserContainer object.
        """

        self.init_containers.append(init_container)
        return self

    def add_sidecar(self, sidecar: Sidecar):
        """Add a sidecar to the Op.

        Args:
          sidecar: SideCar object.
        """

        self.sidecars.append(sidecar)
        return self

    def set_display_name(self, name: str):
        self.display_name = name
        return self

    def set_caching_options(self, enable_caching: bool) -> 'BaseOp':
        """Sets caching options for the Op.

        Args:
          enable_caching: Whether or not to enable caching for this task.

        Returns:
          Self return to allow chained setting calls.
        """
        self.enable_caching = enable_caching
        return self

    def __repr__(self):
        return str({self.__class__.__name__: self.__dict__})


from ._pipeline_volume import \
    PipelineVolume  # The import is here to prevent circular reference problems.


class InputArgumentPath:

    def __init__(self, argument, input=None, path=None):
        self.argument = argument
        self.input = input
        self.path = path


def _is_legacy_output_name(output_name: str) -> Tuple[bool, str]:
    normalized_output_name = re.sub('[^a-zA-Z0-9]', '-', output_name.lower())
    if normalized_output_name in [
            'mlpipeline-ui-metadata',
            'mlpipeline-metrics',
    ]:
        return True, normalized_output_name
    return False, output_name


class ContainerOp(BaseOp):
    """Represents an op implemented by a container image.

    Args:
      name: the name of the op. It does not have to be unique within a
        pipeline because the pipeline will generates a unique new name in case
        of conflicts.
      image: the container image name, such as 'python:3.5-jessie'
      command: the command to run in the container. If None, uses default CMD
        in defined in container.
      arguments: the arguments of the command. The command can include "%s"
        and supply a PipelineParam as the string replacement. For example,
        ('echo %s' % input_param). At container run time the argument will be
        'echo param_value'.
      init_containers: the list of `UserContainer` objects describing the
        InitContainer to deploy before the `main` container.
      sidecars: the list of `Sidecar` objects describing the sidecar
        containers to deploy together with the `main` container.
      container_kwargs: the dict of additional keyword arguments to pass to
        the op's `Container` definition.
      artifact_argument_paths: Optional. Maps input artifact arguments (values
        or references) to the local file paths where they'll be placed. At
        pipeline run time, the value of the artifact argument is saved to a
        local file with specified path. This parameter is only needed when the
        input file paths are hard-coded in the program. Otherwise it's better
        to pass input artifact placement paths by including artifact arguments
        in the command-line using the InputArgumentPath class instances.
      file_outputs: Maps output names to container local output file paths.
        The system will take the data from those files and will make it
        available for passing to downstream tasks. For each output in the
        file_outputs map there will be a corresponding output reference
        available in the task.outputs dictionary. These output references can
        be passed to the other tasks as arguments. The following output names
        are handled specially by the frontend and
            backend: "mlpipeline-ui-metadata" and "mlpipeline-metrics".
      output_artifact_paths: Deprecated. Maps output artifact labels to local
        artifact file paths. Deprecated: Use file_outputs instead. It now
          supports big data outputs.
      is_exit_handler: Deprecated. This is no longer needed.
      pvolumes: Dictionary for the user to match a path on the op's fs with a
        V1Volume or it inherited type.
          E.g {"/my/path": vol, "/mnt": other_op.pvolumes["/output"]}.

    Example::

      from kfp import dsl
      from kubernetes.client.models import V1EnvVar, V1SecretKeySelector
      @dsl.pipeline(
          name='foo',
          description='hello world')
      def foo_pipeline(tag: str, pull_image_policy: str):
        # any attributes can be parameterized (both serialized string or actual PipelineParam)
        op = dsl.ContainerOp(name='foo',
                            image='busybox:%s' % tag,
                            # pass in init_container list
                            init_containers=[dsl.UserContainer('print', 'busybox:latest', command='echo "hello"')],
                            # pass in sidecars list
                            sidecars=[dsl.Sidecar('print', 'busybox:latest', command='echo "hello"')],
                            # pass in k8s container kwargs
                            container_kwargs={'env': [V1EnvVar('foo', 'bar')]},
        )
        # set `imagePullPolicy` property for `container` with `PipelineParam`
        op.container.set_image_pull_policy(pull_image_policy)
        # add sidecar with parameterized image tag
        # sidecar follows the argo sidecar swagger spec
        op.add_sidecar(dsl.Sidecar('redis', 'redis:%s' % tag).set_image_pull_policy('Always'))
    """

    # list of attributes that might have pipeline params - used to generate
    # the input parameters during compilation.
    # Excludes `file_outputs` and `outputs` as they are handled separately
    # in the compilation process to generate the DAGs and task io parameters.

    _DISABLE_REUSABLE_COMPONENT_WARNING = False

    def __init__(
        self,
        name: str,
        image: str,
        command: Optional[StringOrStringList] = None,
        arguments: Optional[ArgumentOrArguments] = None,
        init_containers: Optional[List[UserContainer]] = None,
        sidecars: Optional[List[Sidecar]] = None,
        container_kwargs: Optional[Dict] = None,
        artifact_argument_paths: Optional[List[InputArgumentPath]] = None,
        file_outputs: Optional[Dict[str, str]] = None,
        output_artifact_paths: Optional[Dict[str, str]] = None,
        is_exit_handler: bool = False,
        pvolumes: Optional[Dict[str, V1Volume]] = None,
    ):
        super().__init__(
            name=name,
            init_containers=init_containers,
            sidecars=sidecars,
            is_exit_handler=is_exit_handler)

        self.attrs_with_pipelineparams = BaseOp.attrs_with_pipelineparams + [
            '_container', 'artifact_arguments', '_parameter_arguments'
        ]  #Copying the BaseOp class variable!

        input_artifact_paths = {}
        artifact_arguments = {}
        file_outputs = dict(file_outputs or {})  # Making a copy
        output_artifact_paths = dict(output_artifact_paths or
                                     {})  # Making a copy

        self._is_v2 = False

        def resolve_artifact_argument(artarg):
            from ..components._components import _generate_input_file_name
            if not isinstance(artarg, InputArgumentPath):
                return artarg
            input_name = getattr(
                artarg.input, 'name',
                artarg.input) or ('input-' + str(len(artifact_arguments)))
            input_path = artarg.path or _generate_input_file_name(input_name)
            input_artifact_paths[input_name] = input_path
            artifact_arguments[input_name] = str(artarg.argument)
            return input_path

        for artarg in artifact_argument_paths or []:
            resolve_artifact_argument(artarg)

        if isinstance(command, Sequence) and not isinstance(command, str):
            command = list(map(resolve_artifact_argument, command))
        if isinstance(arguments, Sequence) and not isinstance(arguments, str):
            arguments = list(map(resolve_artifact_argument, arguments))

        # convert to list if not a list
        command = as_string_list(command)
        arguments = as_string_list(arguments)

        if (not ContainerOp._DISABLE_REUSABLE_COMPONENT_WARNING) and (
                '--component_launcher_class_path' not in (arguments or [])):
            # The warning is suppressed for pipelines created using the TFX SDK.
            warnings.warn(
                'Please create reusable components instead of constructing ContainerOp instances directly.'
                ' Reusable components are shareable, portable and have compatibility and support guarantees.'
                ' Please see the documentation: https://www.kubeflow.org/docs/pipelines/sdk/component-development/#writing-your-component-definition-file'
                ' The components can be created manually (or, in case of python, using kfp.components.create_component_from_func or func_to_container_op)'
                ' and then loaded using kfp.components.load_component_from_file, load_component_from_uri or load_component_from_text: '
                'https://kubeflow-pipelines.readthedocs.io/en/stable/source/kfp.components.html#kfp.components.load_component_from_file',
                category=FutureWarning,
            )
            if COMPILING_FOR_V2:
                raise RuntimeError(
                    'Constructing ContainerOp instances directly is deprecated and not '
                    'supported when compiling to v2 (using v2 compiler or v1 compiler '
                    'with V2_COMPATIBLE or V2_ENGINE mode).')

        # `container` prop in `io.argoproj.workflow.v1alpha1.Template`
        container_kwargs = container_kwargs or {}
        self._container = Container(
            image=image, args=arguments, command=command, **container_kwargs)

        # NOTE for backward compatibility (remove in future?)
        # proxy old ContainerOp callables to Container

        # attributes to NOT proxy
        ignore_set = frozenset(['to_dict', 'to_str'])

        # decorator func to proxy a method in `Container` into `ContainerOp`
        def _proxy(proxy_attr):
            """Decorator func to proxy to ContainerOp.container."""

            def _decorated(*args, **kwargs):
                # execute method
                ret = getattr(self._container, proxy_attr)(*args, **kwargs)
                if ret == self._container:
                    return self
                return ret

            return deprecation_warning(_decorated, proxy_attr, proxy_attr)

        # iter thru container and attach a proxy func to the container method
        for attr_to_proxy in dir(self._container):
            func = getattr(self._container, attr_to_proxy)
            # ignore private methods, and bypass method overrided by subclasses.
            if (not hasattr(self, attr_to_proxy) and
                    hasattr(func, '__call__') and (attr_to_proxy[0] != '_') and
                (attr_to_proxy not in ignore_set)):
                # only proxy public callables
                setattr(self, attr_to_proxy, _proxy(attr_to_proxy))

        if output_artifact_paths:
            warnings.warn(
                'The output_artifact_paths parameter is deprecated since SDK v0.1.32. '
                'Use the file_outputs parameter instead. file_outputs now supports '
                'outputting big data.', DeprecationWarning)

        # Skip the special handling that is unnecessary in v2.
        if not COMPILING_FOR_V2:
            # Special handling for the mlpipeline-ui-metadata and mlpipeline-metrics
            # outputs that should always be saved as artifacts
            # TODO: Remove when outputs are always saved as artifacts
            for output_name, path in dict(file_outputs).items():
                is_legacy_name, normalized_name = _is_legacy_output_name(
                    output_name)
                if is_legacy_name:
                    output_artifact_paths[normalized_name] = path
                    del file_outputs[output_name]

        # attributes specific to `ContainerOp`
        self.input_artifact_paths = input_artifact_paths
        self.artifact_arguments = artifact_arguments
        self.file_outputs = file_outputs
        self.output_artifact_paths = output_artifact_paths or {}

        self._metadata = None
        self._parameter_arguments = None

        self.execution_options = _structures.ExecutionOptionsSpec(
            caching_strategy=_structures.CachingStrategySpec(),)

        self.outputs = {}
        if file_outputs:
            self.outputs = {
                name: _pipeline_param.PipelineParam(name, op_name=self.name)
                for name in file_outputs.keys()
            }

        self._set_single_output_attribute()

        self.pvolumes = {}
        self.add_pvolumes(pvolumes)

    def _set_single_output_attribute(self):
        # Syntactic sugar: Add task.output attribute if the component has a single
        # output.
        # TODO: Currently the "MLPipeline UI Metadata" output is removed from
        # outputs to preserve backwards compatibility. Maybe stop excluding it from
        # outputs, but rather exclude it from unique_outputs.
        unique_outputs = set(self.outputs.values())
        if len(unique_outputs) == 1:
            self.output = list(unique_outputs)[0]
        else:
            self.output = _MultipleOutputsError()

    @property
    def is_v2(self):
        return self._is_v2

    @is_v2.setter
    def is_v2(self, is_v2: bool):
        self._is_v2 = is_v2

    # v2 container spec
    @property
    def container_spec(self):
        return self._container._container_spec

    @container_spec.setter
    def container_spec(self, spec: _PipelineContainerSpec):
        if not isinstance(spec, _PipelineContainerSpec):
            raise TypeError('container_spec can only be PipelineContainerSpec. '
                            'Got: {}'.format(spec))
        self._container._container_spec = spec

    @property
    def command(self):
        return self._container.command

    @command.setter
    def command(self, value):
        self._container.command = as_string_list(value)

    @property
    def arguments(self):
        return self._container.args

    @arguments.setter
    def arguments(self, value):
        self._container.args = as_string_list(value)

    @property
    def container(self):
        """`Container` object that represents the `container` property in
        `io.argoproj.workflow.v1alpha1.Template`. Can be used to update the
        container configurations.

        Example::

          import kfp.dsl as dsl
          from kubernetes.client.models import V1EnvVar

          @dsl.pipeline(name='example_pipeline')
          def immediate_value_pipeline():
            op1 = (dsl.ContainerOp(name='example', image='nginx:alpine')
                    .container
                        .add_env_variable(V1EnvVar(name='HOST',
                        value='foo.bar'))
                        .add_env_variable(V1EnvVar(name='PORT', value='80'))
                        .parent # return the parent `ContainerOp`
                    )
        """
        return self._container

    def _set_metadata(self,
                      metadata,
                      arguments: Optional[Dict[str, Any]] = None):
        """Passes the ContainerOp the metadata information and configures the
        right output.

        Args:
          metadata (ComponentSpec): component metadata
          arguments: Dictionary of input arguments to the component.
        """
        if not isinstance(metadata, _structures.ComponentSpec):
            raise ValueError('_set_metadata is expecting ComponentSpec.')

        self._metadata = metadata

        if self._metadata.outputs:
            declared_outputs = {
                output.name:
                _pipeline_param.PipelineParam(output.name, op_name=self.name)
                for output in self._metadata.outputs
            }
            self.outputs.update(declared_outputs)

            for output in self._metadata.outputs:
                if output.name in self.file_outputs:
                    continue
                is_legacy_name, normalized_name = _is_legacy_output_name(
                    output.name)
                if is_legacy_name and normalized_name in self.output_artifact_paths:
                    output_filename = self.output_artifact_paths[
                        normalized_name]
                else:
                    output_filename = _components._generate_output_file_name(
                        output.name)
                self.file_outputs[output.name] = output_filename

            if not COMPILING_FOR_V2:
                for output_name, path in dict(self.file_outputs).items():
                    is_legacy_name, normalized_name = _is_legacy_output_name(
                        output_name)
                    if is_legacy_name:
                        self.output_artifact_paths[normalized_name] = path
                        del self.file_outputs[output_name]
                        del self.outputs[output_name]

        if arguments is not None:
            for input_name, value in arguments.items():
                self.artifact_arguments[input_name] = str(value)
                if (isinstance(value, _pipeline_param.PipelineParam)):
                    self._component_spec_inputs_with_pipeline_params.append(
                        value)

                if input_name not in self.input_artifact_paths:
                    input_artifact_path = _components._generate_input_file_name(
                        input_name)
                    self.input_artifact_paths[input_name] = input_artifact_path

        if self.file_outputs:
            for output in self.file_outputs.keys():
                output_type = self.outputs[output].param_type
                for output_meta in self._metadata.outputs:
                    if output_meta.name == output:
                        output_type = output_meta.type
                self.outputs[output].param_type = output_type

        self._set_single_output_attribute()

    def add_pvolumes(self, pvolumes: Dict[str, V1Volume] = None):
        """Updates the existing pvolumes dict, extends volumes and
        volume_mounts and redefines the pvolume attribute.

        Args:
          pvolumes: Dictionary. Keys are mount paths, values are Kubernetes
            volumes or inherited types (e.g. PipelineVolumes).
        """
        if pvolumes:
            for mount_path, pvolume in pvolumes.items():
                if hasattr(pvolume, 'dependent_names'):
                    self.dependent_names.extend(pvolume.dependent_names)
                else:
                    pvolume = PipelineVolume(volume=pvolume)
                pvolume = pvolume.after(self)
                self.pvolumes[mount_path] = pvolume
                self.add_volume(pvolume)
                self._container.add_volume_mount(
                    V1VolumeMount(name=pvolume.name, mount_path=mount_path))

        self.pvolume = None
        if len(self.pvolumes) == 1:
            self.pvolume = list(self.pvolumes.values())[0]
        return self

    def add_node_selector_constraint(
            self, label_name: Union[str, _pipeline_param.PipelineParam],
            value: Union[str, _pipeline_param.PipelineParam]) -> 'ContainerOp':
        """Sets accelerator type requirement for this task.

        When compiling for v2, this function can be optionally used with
        set_gpu_limit to set the number of accelerator required. Otherwise, by
        default the number requested will be 1.

        Args:
          label_name: The name of the constraint label.
            For v2, only 'cloud.google.com/gke-accelerator' is supported now.
          value: The name of the accelerator.
            For v2, available values include 'nvidia-tesla-k80', 'tpu-v3'.

        Returns:
          self return to allow chained call with other resource specification.
        """
        if self.container_spec and not (
                isinstance(label_name, _pipeline_param.PipelineParam) or
                isinstance(value, _pipeline_param.PipelineParam)):
            accelerator_cnt = 1
            if self.container_spec.resources.accelerator.count > 1:
                # Reserve the number if already set.
                accelerator_cnt = self.container_spec.resources.accelerator.count

            accelerator_config = _PipelineContainerSpec.ResourceSpec.AcceleratorConfig(
                type=_sanitize_gpu_type(value), count=accelerator_cnt)
            self.container_spec.resources.accelerator.CopyFrom(
                accelerator_config)

        super(ContainerOp, self).add_node_selector_constraint(label_name, value)
        return self


# proxy old ContainerOp properties to ContainerOp.container
# with PendingDeprecationWarning.
ContainerOp = _proxy_container_op_props(ContainerOp)


class _MultipleOutputsError:

    @staticmethod
    def raise_error():
        raise RuntimeError(
            'This task has multiple outputs. Use `task.outputs[<output name>]` '
            'dictionary to refer to the one you need.')

    def __getattribute__(self, name):
        _MultipleOutputsError.raise_error()

    def __str__(self):
        _MultipleOutputsError.raise_error()


def _get_cpu_number(cpu_string: str) -> float:
    """Converts the cpu string to number of vCPU core."""
    # dsl.ContainerOp._validate_cpu_string guaranteed that cpu_string is either
    # 1) a string can be converted to a float; or
    # 2) a string followed by 'm', and it can be converted to a float.
    if cpu_string.endswith('m'):
        return float(cpu_string[:-1]) / 1000
    else:
        return float(cpu_string)


def _get_resource_number(resource_string: str) -> float:
    """Converts the resource string to number of resource in GB."""
    # dsl.ContainerOp._validate_size_string guaranteed that memory_string
    # represents an integer, optionally followed by one of (E, Ei, P, Pi, T, Ti,
    # G, Gi, M, Mi, K, Ki).
    # See the meaning of different suffix at
    # https://kubernetes.io/docs/concepts/configuration/manage-resources-containers/#meaning-of-memory
    # Also, ResourceSpec in pipeline IR expects a number in GB.
    if resource_string.endswith('E'):
        return float(resource_string[:-1]) * _E / _G
    elif resource_string.endswith('Ei'):
        return float(resource_string[:-2]) * _EI / _G
    elif resource_string.endswith('P'):
        return float(resource_string[:-1]) * _P / _G
    elif resource_string.endswith('Pi'):
        return float(resource_string[:-2]) * _PI / _G
    elif resource_string.endswith('T'):
        return float(resource_string[:-1]) * _T / _G
    elif resource_string.endswith('Ti'):
        return float(resource_string[:-2]) * _TI / _G
    elif resource_string.endswith('G'):
        return float(resource_string[:-1])
    elif resource_string.endswith('Gi'):
        return float(resource_string[:-2]) * _GI / _G
    elif resource_string.endswith('M'):
        return float(resource_string[:-1]) * _M / _G
    elif resource_string.endswith('Mi'):
        return float(resource_string[:-2]) * _MI / _G
    elif resource_string.endswith('K'):
        return float(resource_string[:-1]) * _K / _G
    elif resource_string.endswith('Ki'):
        return float(resource_string[:-2]) * _KI / _G
    else:
        # By default interpret as a plain integer, in the unit of Bytes.
        return float(resource_string) / _G


def _sanitize_gpu_type(gpu_type: str) -> str:
    """Converts the GPU type to conform the enum style."""
    return gpu_type.replace('-', '_').upper()



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for kfp.dsl.metrics_utils."""

import os
import unittest
import json

from kfp.deprecated.dsl import metrics_utils


class MetricsUtilsTest(unittest.TestCase):

    def test_confusion_matrix(self):
        conf_matrix = metrics_utils.ConfusionMatrix()
        conf_matrix.set_categories(['dog', 'cat', 'horses'])
        conf_matrix.log_row('dog', [2, 6, 0])
        conf_matrix.log_cell('cat', 'dog', 3)
        with open(
                os.path.join(
                    os.path.dirname(__file__), 'test_data',
                    'expected_confusion_matrix.json')) as json_file:
            expected_json = json.load(json_file)
            self.assertEqual(expected_json, conf_matrix.get_metrics())

    def test_bulkload_confusion_matrix(self):
        conf_matrix = metrics_utils.ConfusionMatrix()
        conf_matrix.load_matrix(['dog', 'cat', 'horses'],
                                [[2, 6, 0], [3, 5, 6], [5, 7, 8]])

        with open(
                os.path.join(
                    os.path.dirname(__file__), 'test_data',
                    'expected_bulk_loaded_confusion_matrix.json')) as json_file:
            expected_json = json.load(json_file)
            self.assertEqual(expected_json, conf_matrix.get_metrics())

    def test_confidence_metrics(self):
        confid_metrics = metrics_utils.ConfidenceMetrics()
        confid_metrics.confidenceThreshold = 24.3
        confid_metrics.recall = 24.5
        confid_metrics.falsePositiveRate = 98.4
        expected_dict = {
            'confidenceThreshold': 24.3,
            'recall': 24.5,
            'falsePositiveRate': 98.4
        }
        self.assertEqual(expected_dict, confid_metrics.get_metrics())


if __name__ == '__main__':
    unittest.main()



# Copyright 2018 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
from typing import Optional
from kfp.deprecated import dsl
import kfp.deprecated as kfp
import kfp.deprecated.compiler
import os
import sys

_KF_PIPELINES_COMPILER_MODE_ENV = 'KF_PIPELINES_COMPILER_MODE'


def parse_arguments():
    """Parse command line arguments."""

    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--py', type=str, help='local absolute path to a py file.')
    parser.add_argument(
        '--function',
        type=str,
        help='The name of the function to compile if there are multiple.')
    parser.add_argument(
        '--namespace', type=str, help='The namespace for the pipeline function')
    parser.add_argument(
        '--output',
        type=str,
        required=True,
        help='local path to the output workflow yaml file.')
    parser.add_argument(
        '--disable-type-check',
        action='store_true',
        help='disable the type check, default is enabled.')
    parser.add_argument(
        '--mode',
        type=str,
        help='compiler mode, defaults to V1, can also be V2_COMPATIBLE. You can override the default using env var KF_PIPELINES_COMPILER_MODE.'
    )

    args = parser.parse_args()
    return args


def _compile_pipeline_function(
        pipeline_funcs,
        function_name,
        output_path,
        type_check,
        mode: Optional[dsl.PipelineExecutionMode] = None,
        pipeline_conf: Optional[dsl.PipelineConf] = None):
    if len(pipeline_funcs) == 0:
        raise ValueError(
            'A function with @dsl.pipeline decorator is required in the py file.'
        )

    if len(pipeline_funcs) > 1 and not function_name:
        func_names = [x.__name__ for x in pipeline_funcs]
        raise ValueError(
            'There are multiple pipelines: %s. Please specify --function.' %
            func_names)

    if function_name:
        pipeline_func = next(
            (x for x in pipeline_funcs if x.__name__ == function_name), None)
        if not pipeline_func:
            raise ValueError('The function "%s" does not exist. '
                             'Did you forget @dsl.pipeline decoration?' %
                             function_name)
    else:
        pipeline_func = pipeline_funcs[0]

    kfp.deprecated.compiler.Compiler(mode=mode).compile(pipeline_func,
                                                        output_path, type_check,
                                                        pipeline_conf)


class PipelineCollectorContext():

    def __enter__(self):
        pipeline_funcs = []

        def add_pipeline(func):
            pipeline_funcs.append(func)
            return func

        self.old_handler = dsl._pipeline._pipeline_decorator_handler
        dsl._pipeline._pipeline_decorator_handler = add_pipeline
        return pipeline_funcs

    def __exit__(self, *args):
        dsl._pipeline._pipeline_decorator_handler = self.old_handler


def compile_pyfile(pyfile,
                   output_path,
                   function_name=None,
                   type_check=True,
                   mode: Optional[dsl.PipelineExecutionMode] = None,
                   pipeline_conf: Optional[dsl.PipelineConf] = None):
    sys.path.insert(0, os.path.dirname(pyfile))
    try:
        filename = os.path.basename(pyfile)
        with PipelineCollectorContext() as pipeline_funcs:
            __import__(os.path.splitext(filename)[0])
        _compile_pipeline_function(pipeline_funcs, function_name, output_path,
                                   type_check, mode, pipeline_conf)
    finally:
        del sys.path[0]


def main():
    args = parse_arguments()
    if args.py is None:
        raise ValueError('The --py option must be specified.')
    mode_str = args.mode
    if not mode_str:
        mode_str = os.environ.get(_KF_PIPELINES_COMPILER_MODE_ENV, 'V1')
    mode = None
    if mode_str == 'V1_LEGACY' or mode_str == 'V1':
        mode = kfp.dsl.PipelineExecutionMode.V1_LEGACY
    elif mode_str == 'V2_COMPATIBLE':
        mode = kfp.dsl.PipelineExecutionMode.V2_COMPATIBLE
    elif mode_str == 'V2_ENGINE':
        mode = kfp.dsl.PipelineExecutionMode.V2_ENGINE
    else:
        raise ValueError(
            f'Got unexpected --mode option "{mode_str}", it must be one of V1, V2_COMPATIBLE or V2_ENGINE'
        )
    compile_pyfile(
        args.py,
        args.output,
        args.function,
        not args.disable_type_check,
        mode,
    )



# Copyright 2019 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
from kubernetes import client as k8s_client
from kubernetes import config
import time
import logging
import os


class K8sJobHelper(object):
    """Kubernetes Helper."""

    def __init__(self, k8s_client_configuration=None):
        if not self._configure_k8s(k8s_client_configuration):
            raise Exception('K8sHelper __init__ failure')

    def _configure_k8s(self, k8s_client_configuration=None):
        k8s_config_file = os.environ.get('KUBECONFIG')
        if k8s_config_file:
            try:
                logging.info('Loading kubernetes config from the file %s',
                             k8s_config_file)
                config.load_kube_config(
                    config_file=k8s_config_file,
                    client_configuration=k8s_client_configuration)
            except Exception as e:
                raise RuntimeError(
                    'Can not load kube config from the file %s, error: %s',
                    k8s_config_file, e)
        else:
            try:
                config.load_incluster_config()
                logging.info('Initialized with in-cluster config.')
            except:
                logging.info(
                    'Cannot find in-cluster config, trying the local kubernetes config. '
                )
                try:
                    config.load_kube_config(
                        client_configuration=k8s_client_configuration)
                    logging.info(
                        'Found local kubernetes config. Initialized with kube_config.'
                    )
                except:
                    raise RuntimeError(
                        'Forgot to run the gcloud command? Check out the link: \
          https://cloud.google.com/kubernetes-engine/docs/how-to/cluster-access-for-kubectl for more information'
                    )
        self._api_client = k8s_client.ApiClient()
        self._corev1 = k8s_client.CoreV1Api(self._api_client)
        return True

    def _create_k8s_job(self, yaml_spec):
        """_create_k8s_job creates a kubernetes job based on the yaml spec."""
        pod = k8s_client.V1Pod(
            metadata=k8s_client.V1ObjectMeta(
                generate_name=yaml_spec['metadata']['generateName'],
                annotations=yaml_spec['metadata']['annotations']))
        container = k8s_client.V1Container(
            name=yaml_spec['spec']['containers'][0]['name'],
            image=yaml_spec['spec']['containers'][0]['image'],
            args=yaml_spec['spec']['containers'][0]['args'])
        pod.spec = k8s_client.V1PodSpec(
            restart_policy=yaml_spec['spec']['restartPolicy'],
            containers=[container],
            service_account_name=yaml_spec['spec']['serviceAccountName'])
        try:
            api_response = self._corev1.create_namespaced_pod(
                yaml_spec['metadata']['namespace'], pod)
            return api_response.metadata.name, True
        except k8s_client.rest.ApiException as e:
            logging.exception(
                "Exception when calling CoreV1Api->create_namespaced_pod: {}\n"
                .format(str(e)))
            return '', False

    def _wait_for_k8s_job(self, pod_name, yaml_spec, timeout):
        """_wait_for_k8s_job waits for the job to complete."""
        status = 'running'
        start_time = datetime.now()
        while status in ['pending', 'running']:
            # Pod pending values: https://github.com/kubernetes-client/python/blob/master/kubernetes/docs/V1PodStatus.md
            try:
                api_response = self._corev1.read_namespaced_pod(
                    pod_name, yaml_spec['metadata']['namespace'])
                status = api_response.status.phase.lower()
                time.sleep(5)
                elapsed_time = (datetime.now() - start_time).seconds
                logging.info('{} seconds: waiting for job to complete'.format(
                    elapsed_time))
                if elapsed_time > timeout:
                    logging.info('Kubernetes job timeout')
                    return False
            except k8s_client.rest.ApiException as e:
                logging.exception(
                    'Exception when calling CoreV1Api->read_namespaced_pod: {}\n'
                    .format(str(e)))
                return False
        return status == 'succeeded'

    def _delete_k8s_job(self, pod_name, yaml_spec):
        """_delete_k8s_job deletes a pod."""
        try:
            api_response = self._corev1.delete_namespaced_pod(
                pod_name,
                yaml_spec['metadata']['namespace'],
                body=k8s_client.V1DeleteOptions())
        except k8s_client.rest.ApiException as e:
            logging.exception(
                'Exception when calling CoreV1Api->delete_namespaced_pod: {}\n'
                .format(str(e)))

    def _read_pod_log(self, pod_name, yaml_spec):
        try:
            api_response = self._corev1.read_namespaced_pod_log(
                pod_name, yaml_spec['metadata']['namespace'])
        except k8s_client.rest.ApiException as e:
            logging.exception(
                'Exception when calling CoreV1Api->read_namespaced_pod_log: {}\n'
                .format(str(e)))
            return False
        return api_response

    def _read_pod_status(self, pod_name, namespace):
        try:
            # Using read_namespaced_pod due to the following error: "pods \"kaniko-p2phh\" is forbidden: User \"system:serviceaccount:kubeflow:jupyter-notebook\" cannot get pods/status in the namespace \"kubeflow\""
            #api_response = self._corev1.read_namespaced_pod_status(pod_name, namespace)
            api_response = self._corev1.read_namespaced_pod(pod_name, namespace)
        except k8s_client.rest.ApiException as e:
            logging.exception(
                'Exception when calling CoreV1Api->read_namespaced_pod_status: {}\n'
                .format(str(e)))
            return False
        return api_response

    def run_job(self, yaml_spec, timeout=600):
        """run_job runs a kubernetes job and clean up afterwards."""
        pod_name, succ = self._create_k8s_job(yaml_spec)
        namespace = yaml_spec['metadata']['namespace']
        if not succ:
            raise RuntimeError('Kubernetes job creation failed.')
        # timeout in seconds
        succ = self._wait_for_k8s_job(pod_name, yaml_spec, timeout)
        if not succ:
            logging.info('Kubernetes job failed.')
            print(self._read_pod_log(pod_name, yaml_spec))
            raise RuntimeError('Kubernetes job failed.')
        status_obj = self._read_pod_status(pod_name, namespace)
        self._delete_k8s_job(pod_name, yaml_spec)
        return status_obj



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for handling Python type annotation."""

import re
from typing import TypeVar, Union

T = TypeVar('T')


def maybe_strip_optional_from_annotation(annotation: T) -> T:
    """Strips 'Optional' from 'Optional[<type>]' if applicable.

    For example::
      Optional[str] -> str
      str -> str
      List[int] -> List[int]

    Args:
      annotation: The original type annotation which may or may not has
        `Optional`.

    Returns:
      The type inside Optional[] if Optional exists, otherwise the original type.
    """
    if getattr(annotation, '__origin__',
               None) is Union and annotation.__args__[1] is type(None):
        return annotation.__args__[0]
    return annotation


def get_short_type_name(type_name: str) -> str:
    """Extracts the short form type name.

    This method is used for looking up serializer for a given type.

    For example::
      typing.List -> List
      typing.List[int] -> List
      typing.Dict[str, str] -> Dict
      List -> List
      str -> str

    Args:
      type_name: The original type name.

    Returns:
      The short form type name or the original name if pattern doesn't match.
    """
    match = re.match('(typing\.)?(?P<type>\w+)(?:\[.+\])?', type_name)
    if match:
        return match.group('type')
    else:
        return type_name



# Copyright 2020-2022 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility function for building Importer Node spec."""

from typing import Any, Dict, Mapping, Optional, Type, Union

from kfp.components import importer_component
from kfp.components import pipeline_channel
from kfp.components import pipeline_task
from kfp.components import placeholders
from kfp.components import structures
from kfp.components import utils
from kfp.components.types import artifact_types
from kfp.components.types import type_utils

URI_KEY = 'uri'
OUTPUT_KEY = 'artifact'
METADATA_KEY = 'metadata'


def importer(
    artifact_uri: Union[pipeline_channel.PipelineParameterChannel, str],
    artifact_class: Type[artifact_types.Artifact],
    reimport: bool = False,
    metadata: Optional[Mapping[str, Any]] = None,
) -> pipeline_task.PipelineTask:
    """Imports an existing artifact for use in a downstream component.

    Args:
      artifact_uri: The URI of the artifact to import.
      artifact_class: The artifact class being imported.
      reimport: Whether to reimport the artifact.
      metadata: Properties of the artifact.

    Returns:
      A task with the artifact accessible via its ``.output`` attribute.

    Examples::

      @dsl.pipeline(name='pipeline-with-importer')
      def pipeline_with_importer():

          importer1 = importer(
              artifact_uri='gs://ml-pipeline-playground/shakespeare1.txt',
              artifact_class=Dataset,
              reimport=False)
          train(dataset=importer1.output)
    """

    component_inputs: Dict[str, structures.InputSpec] = {}
    call_inputs: Dict[str, Any] = {}

    def traverse_dict_and_create_metadata_inputs(d: Any) -> Any:
        if isinstance(d, pipeline_channel.PipelineParameterChannel):
            reversed_call_inputs = {
                pipeline_param_chan: name
                for name, pipeline_param_chan in call_inputs.items()
            }

            # minimizes importer spec interface by not creating new
            # inputspec/parameters if the same input is used multiple places
            # in metadata
            unique_name = reversed_call_inputs.get(
                d,
                utils.make_name_unique_by_adding_index(
                    METADATA_KEY,
                    list(call_inputs),
                    '-',
                ),
            )

            call_inputs[unique_name] = d
            component_inputs[unique_name] = structures.InputSpec(
                type=d.channel_type)

            return placeholders.InputValuePlaceholder(
                input_name=unique_name)._to_string()

        elif isinstance(d, dict):
            # use this instead of list comprehension to ensure compiles are identical across Python versions
            res = {}
            for k, v in d.items():
                new_k = traverse_dict_and_create_metadata_inputs(k)
                new_v = traverse_dict_and_create_metadata_inputs(v)
                res[new_k] = new_v
            return res

        elif isinstance(d, list):
            return [traverse_dict_and_create_metadata_inputs(el) for el in d]

        elif isinstance(d, str):
            # extract pipeline channels from f-strings, if any
            pipeline_channels = pipeline_channel.extract_pipeline_channels_from_any(
                d)

            # pass the channel back into the recursive function to create the placeholder, component inputs, and call inputs, then replace the channel with the placeholder
            for channel in pipeline_channels:
                input_placeholder = traverse_dict_and_create_metadata_inputs(
                    channel)
                d = d.replace(channel.pattern, input_placeholder)
            return d

        else:
            return d

    metadata_with_placeholders = traverse_dict_and_create_metadata_inputs(
        metadata)

    component_spec = structures.ComponentSpec(
        name='importer',
        implementation=structures.Implementation(
            importer=structures.ImporterSpec(
                artifact_uri=placeholders.InputValuePlaceholder(
                    URI_KEY)._to_string(),
                schema_title=type_utils.create_bundled_artifact_type(
                    artifact_class.schema_title, artifact_class.schema_version),
                schema_version=artifact_class.schema_version,
                reimport=reimport,
                metadata=metadata_with_placeholders)),
        inputs={
            URI_KEY: structures.InputSpec(type='String'),
            **component_inputs
        },
        outputs={
            OUTPUT_KEY:
                structures.OutputSpec(
                    type=type_utils.create_bundled_artifact_type(
                        artifact_class.schema_title,
                        artifact_class.schema_version))
        },
    )
    importer = importer_component.ImporterComponent(
        component_spec=component_spec)
    return importer(uri=artifact_uri, **call_inputs)



# Copyright 2020 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for kfp.dsl.serialization_utils module."""
import unittest

from kfp.deprecated.dsl import serialization_utils

_DICT_DATA = {
    'int1': 1,
    'str1': 'helloworld',
    'float1': 1.11,
    'none1': None,
    'dict1': {
        'int2': 2,
        'list2': ['inside the list', None, 42]
    }
}

_EXPECTED_YAML_LITERAL = """\
dict1:
  int2: 2
  list2:
  - inside the list
  -
  - 42
float1: 1.11
int1: 1
none1:
str1: helloworld
"""


class SerializationUtilsTest(unittest.TestCase):

    def testDumps(self):
        self.assertEqual(_EXPECTED_YAML_LITERAL,
                         serialization_utils.yaml_dump(_DICT_DATA))



from typing import Dict, Optional
from kfp.deprecated import dsl


def mount_pvc(pvc_name='pipeline-claim',
              volume_name='pipeline',
              volume_mount_path='/mnt/pipeline'):
    """Modifier function to apply to a Container Op to simplify volume, volume
    mount addition and enable better reuse of volumes, volume claims across
    container ops.

    Example:
        ::

            train = train_op(...)
            train.apply(mount_pvc('claim-name', 'pipeline', '/mnt/pipeline'))
    """

    def _mount_pvc(task):
        from kubernetes import client as k8s_client
        # there can be other ops in a pipeline (e.g. ResourceOp, VolumeOp)
        # refer to #3906
        if not hasattr(task, "add_volume") or not hasattr(
                task, "add_volume_mount"):
            return task
        local_pvc = k8s_client.V1PersistentVolumeClaimVolumeSource(
            claim_name=pvc_name)
        return (task.add_volume(
            k8s_client.V1Volume(
                name=volume_name,
                persistent_volume_claim=local_pvc)).add_volume_mount(
                    k8s_client.V1VolumeMount(
                        mount_path=volume_mount_path, name=volume_name)))

    return _mount_pvc


def use_k8s_secret(
    secret_name: str = 'k8s-secret',
    k8s_secret_key_to_env: Optional[Dict] = None,
):
    """An operator that configures the container to use k8s credentials.

    k8s_secret_key_to_env specifies a mapping from the name of the keys in the k8s secret to the name of the
    environment variables where the values will be added.

    The secret needs to be deployed manually a priori.

    Example:
        ::

            train = train_op(...)
            train.apply(use_k8s_secret(secret_name='s3-secret',
            k8s_secret_key_to_env={'secret_key': 'AWS_SECRET_ACCESS_KEY'}))

        This will load the value in secret 's3-secret' at key 'secret_key' and source it as the environment variable
        'AWS_SECRET_ACCESS_KEY'. I.e. it will produce the following section on the pod:
        env:
        - name: AWS_SECRET_ACCESS_KEY
          valueFrom:
            secretKeyRef:
              name: s3-secret
              key: secret_key
    """

    k8s_secret_key_to_env = k8s_secret_key_to_env or {}

    def _use_k8s_secret(task):
        from kubernetes import client as k8s_client
        for secret_key, env_var in k8s_secret_key_to_env.items():
            task.container \
                .add_env_variable(
                k8s_client.V1EnvVar(
                    name=env_var,
                    value_from=k8s_client.V1EnvVarSource(
                        secret_key_ref=k8s_client.V1SecretKeySelector(
                            name=secret_name,
                            key=secret_key
                        )
                    )
                )
            )
        return task

    return _use_k8s_secret


def add_default_resource_spec(
    memory_limit: Optional[str] = None,
    cpu_limit: Optional[str] = None,
    memory_request: Optional[str] = None,
    cpu_request: Optional[str] = None,
):
    """Add default resource requests and limits.

    For resource units, refer to https://kubernetes.io/docs/concepts/configuration/manage-resources-containers/#resource-units-in-kubernetes.

    Args:
      memory_limit: optional, memory limit. Format can be 512Mi, 2Gi etc.
      cpu_limit: optional, cpu limit. Format can be 0.5, 500m etc.
      memory_request: optional, defaults to memory limit.
      cpu_request: optional, defaults to cpu limit.
    """
    if not memory_request:
        memory_request = memory_limit
    if not cpu_request:
        cpu_request = cpu_limit

    def _add_default_resource_spec(task):
        # Skip tasks which are not container ops.
        if not isinstance(task, dsl.ContainerOp):
            return task
        _apply_default_resource(task, 'cpu', cpu_request, cpu_limit)
        _apply_default_resource(task, 'memory', memory_request, memory_limit)
        return task

    return _add_default_resource_spec


def _apply_default_resource(task: dsl.ContainerOp, resource_name: str,
                            default_request: Optional[str],
                            default_limit: Optional[str]):
    if task.container.get_resource_limit(resource_name):
        # Do nothing.
        # Limit is set, request will default to limit if not set (Kubernetes default behavior),
        # so we do not need to further apply defaults.
        return

    if default_limit:
        task.container.add_resource_limit(resource_name, default_limit)
    if default_request:
        if not task.container.get_resource_request(resource_name):
            task.container.add_resource_request(resource_name, default_request)



# Copyright 2019 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click
import json
from typing import List, Optional

import kfp_server_api
from kfp.deprecated.cli.output import print_output, OutputFormat


@click.group()
def pipeline():
    """Manage pipeline resources."""


@pipeline.command()
@click.option("-p", "--pipeline-name", help="Name of the pipeline.")
@click.option("-d", "--description", help="Description for the pipeline.")
@click.argument("package-file")
@click.pass_context
def upload(ctx: click.Context,
           pipeline_name: str,
           package_file: str,
           description: str = None):
    """Upload a KFP pipeline."""
    client = ctx.obj["client"]
    output_format = ctx.obj["output"]
    if not pipeline_name:
        pipeline_name = package_file.split(".")[0]

    pipeline = client.upload_pipeline(package_file, pipeline_name, description)
    _display_pipeline(pipeline, output_format)


@pipeline.command()
@click.option("-p", "--pipeline-id", help="ID of the pipeline", required=False)
@click.option("-n", "--pipeline-name", help="Name of pipeline", required=False)
@click.option(
    "-v",
    "--pipeline-version",
    help="Name of the pipeline version",
    required=True)
@click.argument("package-file")
@click.pass_context
def upload_version(ctx: click.Context,
                   package_file: str,
                   pipeline_version: str,
                   pipeline_id: Optional[str] = None,
                   pipeline_name: Optional[str] = None):
    """Upload a version of the KFP pipeline."""
    client = ctx.obj["client"]
    output_format = ctx.obj["output"]
    if bool(pipeline_id) == bool(pipeline_name):
        raise ValueError("Need to supply 'pipeline-name' or 'pipeline-id'")
    if pipeline_name is not None:
        pipeline_id = client.get_pipeline_id(name=pipeline_name)
        if pipeline_id is None:
            raise ValueError("Can't find a pipeline with name: %s" %
                             pipeline_name)
    version = client.pipeline_uploads.upload_pipeline_version(
        package_file, name=pipeline_version, pipelineid=pipeline_id)
    _display_pipeline_version(version, output_format)


@pipeline.command()
@click.option(
    '--page-token', default='', help="Token for starting of the page.")
@click.option(
    '-m', '--max-size', default=100, help="Max size of the listed pipelines.")
@click.option(
    '--sort-by',
    default="created_at desc",
    help="Can be '[field_name]', '[field_name] desc'. For example, 'name desc'."
)
@click.option(
    '--filter',
    help=(
        "filter: A url-encoded, JSON-serialized Filter protocol buffer "
        "(see [filter.proto](https://github.com/kubeflow/pipelines/blob/master/backend/api/filter.proto))."
    ))
@click.pass_context
def list(ctx: click.Context, page_token: str, max_size: int, sort_by: str,
         filter: str):
    """List uploaded KFP pipelines."""
    client = ctx.obj["client"]
    output_format = ctx.obj["output"]

    response = client.list_pipelines(
        page_token=page_token,
        page_size=max_size,
        sort_by=sort_by,
        filter=filter)
    if response.pipelines:
        _print_pipelines(response.pipelines, output_format)
    else:
        if output_format == OutputFormat.json.name:
            msg = json.dumps([])
        else:
            msg = "No pipelines found"
        click.echo(msg)


@pipeline.command()
@click.argument("pipeline-id")
@click.option(
    '--page-token', default='', help="Token for starting of the page.")
@click.option(
    '-m',
    '--max-size',
    default=100,
    help="Max size of the listed pipeline versions.")
@click.option(
    '--sort-by',
    default="created_at desc",
    help="Can be '[field_name]', '[field_name] desc'. For example, 'name desc'."
)
@click.option(
    '--filter',
    help=(
        "filter: A url-encoded, JSON-serialized Filter protocol buffer "
        "(see [filter.proto](https://github.com/kubeflow/pipelines/blob/master/backend/api/filter.proto))."
    ))
@click.pass_context
def list_versions(ctx: click.Context, pipeline_id: str, page_token: str,
                  max_size: int, sort_by: str, filter: str):
    """List versions of an uploaded KFP pipeline."""
    client = ctx.obj["client"]
    output_format = ctx.obj["output"]

    response = client.list_pipeline_versions(
        pipeline_id,
        page_token=page_token,
        page_size=max_size,
        sort_by=sort_by,
        filter=filter)
    if response.versions:
        _print_pipeline_versions(response.versions, output_format)
    else:
        if output_format == OutputFormat.json.name:
            msg = json.dumps([])
        else:
            msg = "No pipeline or version found"
        click.echo(msg)


@pipeline.command()
@click.argument("version-id")
@click.pass_context
def delete_version(ctx: click.Context, version_id: str):
    """Delete pipeline version.

    Args:
      version_id: id of the pipeline version.

    Returns:
      Object. If the method is called asynchronously, returns the request thread.

    Throws:
      Exception if pipeline version is not found.
    """
    client = ctx.obj["client"]
    return client.delete_pipeline_version(version_id)


@pipeline.command()
@click.argument("pipeline-id")
@click.pass_context
def get(ctx: click.Context, pipeline_id: str):
    """Get detailed information about an uploaded KFP pipeline."""
    client = ctx.obj["client"]
    output_format = ctx.obj["output"]

    pipeline = client.get_pipeline(pipeline_id)
    _display_pipeline(pipeline, output_format)


@pipeline.command()
@click.argument("pipeline-id")
@click.pass_context
def delete(ctx: click.Context, pipeline_id: str):
    """Delete an uploaded KFP pipeline."""
    client = ctx.obj["client"]

    client.delete_pipeline(pipeline_id)
    click.echo(f"{pipeline_id} is deleted")


def _print_pipelines(pipelines: List[kfp_server_api.ApiPipeline],
                     output_format: OutputFormat):
    headers = ["Pipeline ID", "Name", "Uploaded at"]
    data = [[pipeline.id, pipeline.name,
             pipeline.created_at.isoformat()] for pipeline in pipelines]
    print_output(data, headers, output_format, table_format="grid")


def _print_pipeline_versions(versions: List[kfp_server_api.ApiPipelineVersion],
                             output_format: OutputFormat):
    headers = ["Version ID", "Version name", "Uploaded at", "Pipeline ID"]
    data = [[
        version.id, version.name,
        version.created_at.isoformat(),
        next(rr
             for rr in version.resource_references
             if rr.key.type == kfp_server_api.ApiResourceType.PIPELINE).key.id
    ]
            for version in versions]
    print_output(data, headers, output_format, table_format="grid")


def _display_pipeline(pipeline: kfp_server_api.ApiPipeline,
                      output_format: OutputFormat):
    # Pipeline information
    table = [["Pipeline ID", pipeline.id], ["Name", pipeline.name],
             ["Description", pipeline.description],
             ["Uploaded at", pipeline.created_at.isoformat()],
             ["Version ID", pipeline.default_version.id]]

    # Pipeline parameter details
    headers = ["Parameter Name", "Default Value"]
    data = []
    if pipeline.parameters is not None:
        data = [[param.name, param.value] for param in pipeline.parameters]

    if output_format == OutputFormat.table.name:
        print_output([], ["Pipeline Details"], output_format)
        print_output(table, [], output_format, table_format="plain")
        print_output(data, headers, output_format, table_format="grid")
    elif output_format == OutputFormat.json.name:
        output = dict()
        output["Pipeline Details"] = dict(table)
        params = []
        for item in data:
            params.append(dict(zip(headers, item)))
        output["Pipeline Parameters"] = params
        print_output(output, [], output_format)


def _display_pipeline_version(version: kfp_server_api.ApiPipelineVersion,
                              output_format: OutputFormat):
    pipeline_id = next(
        rr for rr in version.resource_references
        if rr.key.type == kfp_server_api.ApiResourceType.PIPELINE).key.id
    table = [["Pipeline ID", pipeline_id], ["Version name", version.name],
             ["Uploaded at", version.created_at.isoformat()],
             ["Version ID", version.id]]

    if output_format == OutputFormat.table.name:
        print_output([], ["Pipeline Version Details"], output_format)
        print_output(table, [], output_format, table_format="plain")
    elif output_format == OutputFormat.json.name:
        print_output(dict(table), [], output_format)



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import contextlib
import enum
import pathlib
import shutil
import subprocess
import tempfile
from typing import Any, List, Optional

_DOCKER_IS_PRESENT = True
try:
    import docker
except ImportError:
    _DOCKER_IS_PRESENT = False

import typer

import kfp.deprecated as kfp
from kfp.components import component_factory, kfp_config, utils

_REQUIREMENTS_TXT = 'requirements.txt'

_DOCKERFILE = 'Dockerfile'

_DOCKERFILE_TEMPLATE = '''
FROM {base_image}

WORKDIR {component_root_dir}
COPY requirements.txt requirements.txt
RUN pip install --no-cache-dir -r requirements.txt
{maybe_copy_kfp_package}
RUN pip install --no-cache-dir {kfp_package_path}
COPY . .
'''

_DOCKERIGNORE = '.dockerignore'

# Location in which to write out shareable YAML for components.
_COMPONENT_METADATA_DIR = 'component_metadata'

_DOCKERIGNORE_TEMPLATE = '''
{}/
'''.format(_COMPONENT_METADATA_DIR)

# Location at which v2 Python function-based components will stored
# in containerized components.
_COMPONENT_ROOT_DIR = pathlib.Path('/usr/local/src/kfp/components')


@contextlib.contextmanager
def _registered_modules():
    registered_modules = {}
    component_factory.REGISTERED_MODULES = registered_modules
    try:
        yield registered_modules
    finally:
        component_factory.REGISTERED_MODULES = None


class _Engine(str, enum.Enum):
    """Supported container build engines."""
    DOCKER = 'docker'
    KANIKO = 'kaniko'
    CLOUD_BUILD = 'cloudbuild'


app = typer.Typer()


def _info(message: Any):
    info = typer.style('INFO', fg=typer.colors.GREEN)
    typer.echo('{}: {}'.format(info, message))


def _warning(message: Any):
    info = typer.style('WARNING', fg=typer.colors.YELLOW)
    typer.echo('{}: {}'.format(info, message))


def _error(message: Any):
    info = typer.style('ERROR', fg=typer.colors.RED)
    typer.echo('{}: {}'.format(info, message))


class _ComponentBuilder():
    """Helper class for building containerized v2 KFP components."""

    def __init__(
        self,
        context_directory: pathlib.Path,
        kfp_package_path: Optional[pathlib.Path] = None,
        component_filepattern: str = '**/*.py',
    ):
        """ComponentBuilder builds containerized components.

        Args:
            context_directory: Directory containing one or more Python files
            with one or more KFP v2 components.
            kfp_package_path: Path to a pip-installable location for KFP.
                This can either be pointing to KFP SDK root directory located in
                a local clone of the KFP repo, or a git+https location.
                If left empty, defaults to KFP on PyPi.
        """
        self._context_directory = context_directory
        self._dockerfile = self._context_directory / _DOCKERFILE
        self._component_filepattern = component_filepattern
        self._components: List[
            component_factory.component_factory.ComponentInfo] = []

        # This is only set if we need to install KFP from local copy.
        self._maybe_copy_kfp_package = ''

        if kfp_package_path is None:
            self._kfp_package_path = 'kfp=={}'.format(kfp.__version__)
        elif kfp_package_path.is_dir():
            _info('Building KFP package from local directory {}'.format(
                typer.style(str(kfp_package_path), fg=typer.colors.CYAN)))
            temp_dir = pathlib.Path(tempfile.mkdtemp())
            try:
                subprocess.run([
                    'python3',
                    kfp_package_path / 'setup.py',
                    'bdist_wheel',
                    '--dist-dir',
                    str(temp_dir),
                ],
                               cwd=kfp_package_path)
                wheel_files = list(temp_dir.glob('*.whl'))
                if len(wheel_files) != 1:
                    _error('Failed to find built KFP wheel under {}'.format(
                        temp_dir))
                    raise typer.Exit(1)

                wheel_file = wheel_files[0]
                shutil.copy(wheel_file, self._context_directory)
                self._kfp_package_path = wheel_file.name
                self._maybe_copy_kfp_package = 'COPY {wheel_name} {wheel_name}'.format(
                    wheel_name=self._kfp_package_path)
            except subprocess.CalledProcessError as e:
                _error('Failed to build KFP wheel locally:\n{}'.format(e))
                raise typer.Exit(1)
            finally:
                _info('Cleaning up temporary directory {}'.format(temp_dir))
                shutil.rmtree(temp_dir)
        else:
            self._kfp_package_path = kfp_package_path

        _info('Building component using KFP package path: {}'.format(
            typer.style(str(self._kfp_package_path), fg=typer.colors.CYAN)))

        self._context_directory_files = [
            file.name
            for file in self._context_directory.glob('*')
            if file.is_file()
        ]

        self._component_files = [
            file for file in self._context_directory.glob(
                self._component_filepattern) if file.is_file()
        ]

        self._base_image = None
        self._target_image = None
        self._load_components()

    def _load_components(self):
        if not self._component_files:
            _error(
                'No component files found matching pattern `{}` in directory {}'
                .format(self._component_filepattern, self._context_directory))
            raise typer.Exit(1)

        for python_file in self._component_files:
            with _registered_modules() as component_modules:
                module_name = python_file.name[:-len('.py')]
                module_directory = python_file.parent
                utils.load_module(
                    module_name=module_name, module_directory=module_directory)

                formatted_module_file = typer.style(
                    str(python_file), fg=typer.colors.CYAN)
                if not component_modules:
                    _error('No KFP components found in file {}'.format(
                        formatted_module_file))
                    raise typer.Exit(1)

                _info('Found {} component(s) in file {}:'.format(
                    len(component_modules), formatted_module_file))
                for name, component in component_modules.items():
                    _info('{}: {}'.format(name, component))
                    self._components.append(component)

        base_images = set([info.base_image for info in self._components])
        target_images = set([info.target_image for info in self._components])

        if len(base_images) != 1:
            _error('Found {} unique base_image values {}. Components'
                   ' must specify the same base_image and target_image.'.format(
                       len(base_images), base_images))
            raise typer.Exit(1)

        self._base_image = base_images.pop()
        if self._base_image is None:
            _error('Did not find a base_image specified in any of the'
                   ' components. A base_image must be specified in order to'
                   ' build the component.')
            raise typer.Exit(1)
        _info('Using base image: {}'.format(
            typer.style(self._base_image, fg=typer.colors.YELLOW)))

        if len(target_images) != 1:
            _error('Found {} unique target_image values {}. Components'
                   ' must specify the same base_image and'
                   ' target_image.'.format(len(target_images), target_images))
            raise typer.Exit(1)

        self._target_image = target_images.pop()
        if self._target_image is None:
            _error('Did not find a target_image specified in any of the'
                   ' components. A target_image must be specified in order'
                   ' to build the component.')
            raise typer.Exit(1)
        _info('Using target image: {}'.format(
            typer.style(self._target_image, fg=typer.colors.YELLOW)))

    def _maybe_write_file(self,
                          filename: str,
                          contents: str,
                          overwrite: bool = False):
        formatted_filename = typer.style(filename, fg=typer.colors.CYAN)
        if filename in self._context_directory_files:
            _info('Found existing file {} under {}.'.format(
                formatted_filename, self._context_directory))
            if not overwrite:
                _info('Leaving this file untouched.')
                return
            else:
                _warning(
                    'Overwriting existing file {}'.format(formatted_filename))
        else:
            _warning('{} not found under {}. Creating one.'.format(
                formatted_filename, self._context_directory))

        filepath = self._context_directory / filename
        with open(filepath, 'w') as f:
            f.write('# Generated by KFP.\n{}'.format(contents))
        _info('Generated file {}.'.format(filepath))

    def maybe_generate_requirements_txt(self):
        self._maybe_write_file(_REQUIREMENTS_TXT, '')

    def maybe_generate_dockerignore(self):
        self._maybe_write_file(_DOCKERIGNORE, _DOCKERIGNORE_TEMPLATE)

    def write_component_files(self):
        for component_info in self._components:
            filename = (
                component_info.output_component_file or
                component_info.function_name + '.yaml')
            container_filename = (
                self._context_directory / _COMPONENT_METADATA_DIR / filename)
            container_filename.parent.mkdir(exist_ok=True, parents=True)
            component_info.component_spec.save_to_component_yaml(
                str(container_filename))

    def generate_kfp_config(self):
        config = kfp_config.KFPConfig(config_directory=self._context_directory)
        for component_info in self._components:
            relative_path = component_info.module_path.relative_to(
                self._context_directory)
            config.add_component(
                function_name=component_info.function_name, path=relative_path)
        config.save()

    def maybe_generate_dockerfile(self, overwrite_dockerfile: bool = False):
        dockerfile_contents = _DOCKERFILE_TEMPLATE.format(
            base_image=self._base_image,
            maybe_copy_kfp_package=self._maybe_copy_kfp_package,
            component_root_dir=_COMPONENT_ROOT_DIR,
            kfp_package_path=self._kfp_package_path)

        self._maybe_write_file(_DOCKERFILE, dockerfile_contents,
                               overwrite_dockerfile)

    def build_image(self, push_image: bool = True):
        _info('Building image {} using Docker...'.format(
            typer.style(self._target_image, fg=typer.colors.YELLOW)))
        client = docker.from_env()

        docker_log_prefix = typer.style('Docker', fg=typer.colors.CYAN)

        try:
            context = str(self._context_directory)
            logs = client.api.build(
                path=context,
                dockerfile='Dockerfile',
                tag=self._target_image,
                decode=True,
            )
            for log in logs:
                message = log.get('stream', '').rstrip('\n')
                if message:
                    _info('{}: {}'.format(docker_log_prefix, message))

        except docker.errors.BuildError as e:
            for log in e.build_log:
                message = log.get('message', '').rstrip('\n')
                if message:
                    _error('{}: {}'.format(docker_log_prefix, message))
            _error('{}: {}'.format(docker_log_prefix, e))
            raise typer.Exit(1)

        if not push_image:
            return

        _info('Pushing image {}...'.format(
            typer.style(self._target_image, fg=typer.colors.YELLOW)))

        try:
            response = client.images.push(
                self._target_image, stream=True, decode=True)
            for log in response:
                status = log.get('status', '').rstrip('\n')
                layer = log.get('id', '')
                if status:
                    _info('{}: {} {}'.format(docker_log_prefix, layer, status))
        except docker.errors.BuildError as e:
            _error('{}: {}'.format(docker_log_prefix, e))
            raise e

        _info('Built and pushed component container {}'.format(
            typer.style(self._target_image, fg=typer.colors.YELLOW)))


@app.callback()
def components():
    """Builds shareable, containerized components."""


@app.command()
def build(components_directory: pathlib.Path = typer.Argument(
    ...,
    help="Path to a directory containing one or more Python"
    " files with KFP v2 components. The container will be built"
    " with this directory as the context."),
          component_filepattern: str = typer.Option(
              '**/*.py',
              help="Filepattern to use when searching for KFP components. The"
              " default searches all Python files in the specified directory."),
          engine: _Engine = typer.Option(
              _Engine.DOCKER,
              help="Engine to use to build the component's container."),
          kfp_package_path: Optional[pathlib.Path] = typer.Option(
              None, help="A pip-installable path to the KFP package."),
          overwrite_dockerfile: bool = typer.Option(
              False,
              help="Set this to true to always generate a Dockerfile"
              " as part of the build process"),
          push_image: bool = typer.Option(
              True, help="Push the built image to its remote repository.")):
    """Builds containers for KFP v2 Python-based components."""
    components_directory = components_directory.resolve()
    if not components_directory.is_dir():
        _error('{} does not seem to be a valid directory.'.format(
            components_directory))
        raise typer.Exit(1)

    if engine != _Engine.DOCKER:
        _error('Currently, only `docker` is supported for --engine.')
        raise typer.Exit(1)

    if engine == _Engine.DOCKER:
        if not _DOCKER_IS_PRESENT:
            _error(
                'The `docker` Python package was not found in the current'
                ' environment. Please run `pip install docker` to install it.'
                ' Optionally, you can also  install KFP with all of its'
                ' optional dependencies by running `pip install kfp[all]`.')
            raise typer.Exit(1)

    builder = _ComponentBuilder(
        context_directory=components_directory,
        kfp_package_path=kfp_package_path,
        component_filepattern=component_filepattern,
    )
    builder.write_component_files()
    builder.generate_kfp_config()

    builder.maybe_generate_requirements_txt()
    builder.maybe_generate_dockerignore()
    builder.maybe_generate_dockerfile(overwrite_dockerfile=overwrite_dockerfile)
    builder.build_image(push_image=push_image)


if __name__ == '__main__':
    app()



# Copyright 2020 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for serialization/deserialization."""
from typing import Any, Dict
import yaml


# Serialize None as blank instead of 'null'.
def _represent_none(self, _):
    return self.represent_scalar('tag:yaml.org,2002:null', '')


class _NoneAsBlankDumper(yaml.SafeDumper):
    """Alternative dumper to print YAML literal.

    The behavior is mostly identical to yaml.SafeDumper, except for this
    dumper dumps None object as blank, instead of 'null'.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.add_representer(type(None), _represent_none)


def yaml_dump(data: Dict[str, Any]) -> str:
    """Dumps YAML string.

    None will be represented as blank.
    """
    return yaml.dump(data, Dumper=_NoneAsBlankDumper)



# Copyright 2020 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for kfp.dsl.container_op."""
import unittest

from kfp.deprecated.dsl import _container_op
from kfp.pipeline_spec import pipeline_spec_pb2

from google.protobuf import text_format
from google.protobuf import json_format

_EXPECTED_CONTAINER_WITH_RESOURCE = """
resources {
  cpu_limit: 1.0
  memory_limit: 1.0
  accelerator {
    type: 'NVIDIA_TESLA_K80'
    count: 1
  }
}
"""

# Shorthand for PipelineContainerSpec
_PipelineContainerSpec = pipeline_spec_pb2.PipelineDeploymentConfig.PipelineContainerSpec


class ContainerOpTest(unittest.TestCase):

    def test_chained_call_resource_setter(self):
        task = _container_op.ContainerOp(name='test_task', image='python:3.7')
        task.container_spec = _PipelineContainerSpec()
        (task.set_cpu_limit('1').set_memory_limit(
            '1G').add_node_selector_constraint(
                'cloud.google.com/gke-accelerator',
                'nvidia-tesla-k80').set_gpu_limit(1))

        expected_container_spec = text_format.Parse(
            _EXPECTED_CONTAINER_WITH_RESOURCE, _PipelineContainerSpec())

        self.assertDictEqual(
            json_format.MessageToDict(task.container_spec),
            json_format.MessageToDict(expected_container_spec))


if __name__ == '__main__':
    unittest.main()



# Copyright 2022 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Tuple, Union

import click


class AliasedPluralsGroup(click.Group):

    def get_command(self, ctx: click.Context, cmd_name: str) -> click.Command:
        regular = click.Group.get_command(self, ctx, cmd_name)
        if regular is not None:
            return regular
        elif cmd_name.endswith('s'):
            singular = click.Group.get_command(self, ctx, cmd_name[:-1])
            if singular is not None:
                return singular
        raise click.UsageError(f"Unrecognized command '{cmd_name}'")

    def resolve_command(
        self, ctx: click.Context, args: List[str]
    ) -> Tuple[Union[str, None], Union[click.Command, None], List[str]]:
        # always return the full command name
        _, cmd, args = super().resolve_command(ctx, args)
        return cmd.name, cmd, args  # type: ignore



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for kfp.components.pipeline_channel."""

import unittest

from absl.testing import parameterized
from kfp import dsl
from kfp.components import pipeline_channel


class PipelineChannelTest(parameterized.TestCase):

    def test_instantiate_pipline_channel(self):
        with self.assertRaisesRegex(
                TypeError, "Can't instantiate abstract class PipelineChannel"):
            p = pipeline_channel.PipelineChannel(
                name='channel',
                channel_type='String',
            )

    def test_invalid_name(self):
        with self.assertRaisesRegex(
                ValueError,
                'Only letters, numbers, spaces, "_", and "-" are allowed in the '
                'name. Must begin with a letter. Got name: 123_abc'):
            p = pipeline_channel.create_pipeline_channel(
                name='123_abc',
                channel_type='String',
            )

    def test_task_name_and_value_both_set(self):
        with self.assertRaisesRegex(ValueError,
                                    'task_name and value cannot be both set.'):
            p = pipeline_channel.create_pipeline_channel(
                name='abc',
                channel_type='Integer',
                task_name='task1',
                value=123,
            )

    def test_invalid_type(self):
        with self.assertRaisesRegex(TypeError,
                                    'Artifact is not a parameter type.'):
            p = pipeline_channel.PipelineParameterChannel(
                name='channel1',
                channel_type='Artifact',
            )

        with self.assertRaisesRegex(TypeError,
                                    'String is not an artifact type.'):
            p = pipeline_channel.PipelineArtifactChannel(
                name='channel1',
                channel_type='String',
                task_name='task1',
                is_artifact_list=False,
            )

    @parameterized.parameters(
        {
            'pipeline_channel':
                pipeline_channel.create_pipeline_channel(
                    name='channel1',
                    task_name='task1',
                    channel_type='String',
                ),
            'str_repr':
                '{{channel:task=task1;name=channel1;type=String;}}',
        },
        {
            'pipeline_channel':
                pipeline_channel.create_pipeline_channel(
                    name='channel2',
                    channel_type='Integer',
                ),
            'str_repr':
                '{{channel:task=;name=channel2;type=Integer;}}',
        },
        {
            'pipeline_channel':
                pipeline_channel.create_pipeline_channel(
                    name='channel3',
                    channel_type={'type_a': {
                        'property_b': 'c'
                    }},
                    task_name='task3',
                ),
            'str_repr':
                '{{channel:task=task3;name=channel3;type={"type_a": {"property_b": "c"}};}}',
        },
        {
            'pipeline_channel':
                pipeline_channel.create_pipeline_channel(
                    name='channel4',
                    channel_type='Float',
                    value=1.23,
                ),
            'str_repr':
                '{{channel:task=;name=channel4;type=Float;}}',
        },
        {
            'pipeline_channel':
                pipeline_channel.create_pipeline_channel(
                    name='channel5',
                    channel_type='system.Artifact@0.0.1',
                    task_name='task5',
                ),
            'str_repr':
                '{{channel:task=task5;name=channel5;type=system.Artifact@0.0.1;}}',
        },
    )
    def test_str_repr(self, pipeline_channel, str_repr):
        self.assertEqual(str_repr, str(pipeline_channel))

    def test_extract_pipeline_channels(self):
        p1 = pipeline_channel.create_pipeline_channel(
            name='channel1',
            channel_type='String',
            value='abc',
        )
        p2 = pipeline_channel.create_pipeline_channel(
            name='channel2',
            channel_type='customized_type_b',
            task_name='task2',
        )
        p3 = pipeline_channel.create_pipeline_channel(
            name='channel3',
            channel_type={'customized_type_c': {
                'property_c': 'value_c'
            }},
            task_name='task3',
        )
        stuff_chars = ' between '
        payload = str(p1) + stuff_chars + str(p2) + stuff_chars + str(p3)
        params = pipeline_channel.extract_pipeline_channels_from_string(payload)
        self.assertListEqual([p1, p2, p3], params)

        # Expecting the extract_pipelineparam_from_any to dedup pipeline channels
        # among all the payloads.
        payload = [
            str(p1) + stuff_chars + str(p2),
            str(p2) + stuff_chars + str(p3)
        ]
        params = pipeline_channel.extract_pipeline_channels_from_any(payload)
        self.assertListEqual([p1, p2, p3], params)


class TestCanAccessTask(unittest.TestCase):

    def test(self):

        @dsl.component
        def comp() -> str:
            return 'text'

        @dsl.pipeline
        def my_pipeline():
            op1 = comp()
            self.assertEqual(op1.output.task, op1)


if __name__ == '__main__':
    unittest.main()



# Copyright 2021-2022 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""KFP DSL compiler.

Implementation of KFP compiler that compiles KFP pipeline into Pipeline IR:
https://docs.google.com/document/d/1PUDuSQ8vmeKSBloli53mp7GIvzekaY7sggg6ywy35Dk/
"""

from typing import Any, Dict, Optional

from kfp.compiler import pipeline_spec_builder as builder
from kfp.components import base_component
from kfp.components.types import type_utils


class Compiler:
    """Compiles pipelines composed using the KFP SDK DSL to a YAML pipeline
    definition.

    The pipeline definition is `PipelineSpec IR <https://github.com/kubeflow/pipelines/blob/2060e38c5591806d657d85b53eed2eef2e5de2ae/api/v2alpha1/pipeline_spec.proto#L50>`_, the protobuf message that defines a pipeline.

    Example:
      ::

        @dsl.pipeline(
          name='name',
        )
        def my_pipeline(a: int, b: str = 'default value'):
            ...

        kfp.compiler.Compiler().compile(
            pipeline_func=my_pipeline,
            package_path='path/to/pipeline.yaml',
            pipeline_parameters={'a': 1},
        )
    """

    def compile(
        self,
        pipeline_func: base_component.BaseComponent,
        package_path: str,
        pipeline_name: Optional[str] = None,
        pipeline_parameters: Optional[Dict[str, Any]] = None,
        type_check: bool = True,
    ) -> None:
        """Compiles the pipeline or component function into IR YAML.

        Args:
            pipeline_func: Pipeline function constructed with the ``@dsl.pipeline`` or component constructed with the ``@dsl.component`` decorator.
            package_path: Output YAML file path. For example, ``'~/my_pipeline.yaml'`` or ``'~/my_component.yaml'``.
            pipeline_name: Name of the pipeline.
            pipeline_parameters: Map of parameter names to argument values.
            type_check: Whether to enable type checking of component interfaces during compilation.
        """

        with type_utils.TypeCheckManager(enable=type_check):
            if not isinstance(pipeline_func, base_component.BaseComponent):
                raise ValueError(
                    'Unsupported pipeline_func type. Expected '
                    'subclass of `base_component.BaseComponent` or '
                    '`Callable` constructed with @dsl.pipeline '
                    f'decorator. Got: {type(pipeline_func)}')

            pipeline_spec = builder.modify_pipeline_spec_with_override(
                pipeline_spec=pipeline_func.pipeline_spec,
                pipeline_name=pipeline_name,
                pipeline_parameters=pipeline_parameters,
            )

            builder.write_pipeline_spec_to_file(
                pipeline_spec=pipeline_spec,
                pipeline_description=pipeline_func.description,
                platform_spec=pipeline_func.platform_spec,
                package_path=package_path,
            )
